from mammut.curriculum.classroom.single_classroom import SingleClassroom
import ray
from typing import Set
import transitions
import mammut.curriculum.report as report
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


@ray.remote
class RayClassroom(SingleClassroom):
    """Classroom designed to be a workspace in a multi-classroom ray architecture.
    """

    # States
    """
    Closed state means the classroom was closed by Curriculum control, 
    and no new courses has arrived for new processing.
    """
    STATE_CLOSED = "closed"

    """
    Busy state when the classroom is processing a set of courses.
    A classroom will be busy when new courses arrive for processing 
    after being closed.
    """
    STATE_BUSY = "busy"

    """
    Idle state when the classroom isn't processing anything. 
    After processing a set of courses, the classroom will be idle 
    until control client classes close the classroom.
    """
    STATE_IDLE = "idle"

    STATES = [STATE_BUSY, STATE_IDLE, STATE_CLOSED]

    def __init__(self, mammut_session_context: MammutSessionContext):
        """
        Args:
            mammut_session_context: The session context for this classroom and the
                curriculums in it.
        """
        super().__init__(mammut_session_context)
        self.machine = transitions.Machine(
            model=self, states=self.STATES, initial=self.STATE_CLOSED
        )

        # Configure remote reporting. This process worker will report remotely.
        report.set_mammut_session_context(mammut_session_context)
        report.enable_remote()

    def _process_courses_loop(
        self, courses_ids: Set[int], curriculum_package_id: str
    ) -> None:
        """Process the courses specified for the curriculum.

        When processing starts, this classroom will transition to "busy"
        state. After finishing all courses processing, will go to
        "idle" state. Control classes can use the state value to know when
        to fetch new data remotely.

        Args:
            courses_ids(Set[int]): set of courses IDs to be processed
                from the desired curriculum.
            curriculum_package_id: Spreadsheet ID of the curriculum
                to take courses from.
        """
        self.to_busy()
        curriculum = self._curriculums[curriculum_package_id]
        self._last_courses = set(
            [
                course
                for course in curriculum.courses
                if course.course_id in courses_ids
            ]
        )

        report.send_message(
            report.CurriculumGeneralInfoMessage(
                f"Scheduling {len(self._last_courses)} in Classroom."
            )
        )

        report.send_message(
            report.CurriculumGeneralDebugMessage(
                f"Processing {len(self._last_courses)} courses in "
                f"classroom {self.id()}"
            )
        )

        for c in self._last_courses:
            c.start(curriculum.progress, self._mammut_session_context)

        report.send_message(
            report.CurriculumGeneralDebugMessage(
                f"Process completed in classroom: {self.id()}"
            )
        )

        self.to_idle()

    def stop(self):
        """Currently there's nothing to stop """
        pass

    def get_state(self):
        return self.state

    def close(self):
        self.to_closed()



from mammut.curriculum.report.messages.message import (
    CurriculumBaseMessage,
    CurriculumBaseMessageWithLesson,
)
from mammut.curriculum.report.utils import Component
import logging


class CurriculumGeneralInfoMessage(CurriculumBaseMessage):
    """Generic information message for Curriculum processing.
    """

    def __init__(self, message: str):
        super(CurriculumGeneralInfoMessage, self).__init__(
            component=Component.Curriculum,
            report_level=logging.INFO,
            message=message,
        )


class CurriculumGeneralInfoMessageWithLesson(CurriculumBaseMessageWithLesson):
    def __init__(self, message: str, course: int, lesson: int):
        super(CurriculumGeneralInfoMessageWithLesson, self).__init__(
            component=Component.Curriculum,
            course_id=course,
            lesson_id=lesson,
            report_level=logging.INFO,
            message=message,
        )


class CurriculumGeneralDebugMessage(CurriculumBaseMessage):
    """Generic debugging information message for curriculum processing"""

    def __init__(self, message: str):
        super(CurriculumGeneralDebugMessage, self).__init__(
            component=Component.Curriculum,
            report_level=logging.DEBUG,
            message=message,
        )


class CurriculumGeneralErrorMessage(CurriculumBaseMessage):
    """Generic error message for curriculum processing"""

    def __init__(self, message: str):
        super(CurriculumGeneralErrorMessage, self).__init__(
            component=Component.Curriculum,
            report_level=logging.ERROR,
            message=message,
        )


class CurriculumGeneralWarningMessage(CurriculumBaseMessage):
    """Warning general message from curriculum phase.
    """

    def __init__(self, message: str):
        super(CurriculumGeneralWarningMessage, self).__init__(
            component=Component.Curriculum,
            report_level=logging.WARNING,
            message=message,
        )



import os
import sys
from mammut.models.imaging.mammut_image_retraining import *
from ipywidgets import *

### Retrain Image Recognition ###
def get_image_retraining_widget(package, current_mammut_user):

    image_layout = Layout(width="80%", height="50px")
    image_style = {"description_width": "initial"}
    initial_image_message = widgets.HTML(
        value="<b>No package selected</b>",
        placeholder="Message: ",
        description="Message: ",
    )
    samples_textbox = widgets.Text(
        value="50",
        description="Numero de magenes geneneradas aleatoriamente:",
        layout=image_layout,
        style=image_style,
    )
    training_steps_textbox = widgets.Text(
        value="150",
        description="Numero de magenes geneneradas aleatoriamente:",
        layout=image_layout,
        style=image_style,
    )

    def get_retrain_content(package, current_mammut_user):

        images_corpus = ""
        for corpus in package.corpus_map.switch_get[
            package.corpus_map.DEFAULT_IMAGES_TYPE
        ]:
            images_corpus = images_corpus + "<br>" + corpus.sheet_title
        images_corpus_availables = widgets.HTML(
            value=images_corpus,
            placeholder="Images corpus: ",
            description="Images corpus: ",
        )
        retrain_images_button = widgets.Button(
            description="Retrain listed corpus",
            button_style="primary",
            layout=Layout(width="225px", height="100%"),
        )
        retrain_images_content = widgets.VBox(
            [
                images_corpus_availables,
                samples_textbox,
                training_steps_textbox,
                retrain_images_button,
            ]
        )

        def retrain_images_button_clicked(b):
            corpus = package.corpus_map.switch_get[
                package.corpus_map.DEFAULT_IMAGES_TYPE
            ][0]
            mammut_id = current_mammut_user.id
            retrained_graph_path, labels_path = package.image_retraining(
                corpus.destination_folder,
                mammut_id,
                int(samples_textbox.value),
                int(training_steps_textbox.value),
            )
            retrianed_graph_textbox.value = retrained_graph_path
            labels_textbox.value = labels_path

        retrain_images_button.on_click(retrain_images_button_clicked)
        return retrain_images_content

    inference_text = widgets.HTML(
        value="<b>inference</b>", placeholder="Message: ", description="Message: ",
    )

    retrianed_graph_textbox = widgets.Text(
        value="",
        description="retrined graph path:",
        layout=image_layout,
        style=image_style,
    )
    labels_textbox = widgets.Text(
        value="", description="labels path:", layout=image_layout, style=image_style
    )
    test_image_path_text_box = widgets.Text(
        value="", description="test image path:", layout=image_layout, style=image_style
    )

    image_inference_button = widgets.Button(
        description="Probar imagen",
        button_style="primary",
        layout=Layout(width="225px", height="100%"),
    )
    inference_result = widgets.HTML(value="<b></b>")
    images_inference_box = widgets.VBox(
        [
            retrianed_graph_textbox,
            labels_textbox,
            test_image_path_text_box,
            image_inference_button,
            inference_result,
        ]
    )
    retrain_images_box = widgets.VBox(
        [get_retrain_content(package, current_mammut_user), images_inference_box]
    )

    def image_inference_button_clicked(b):
        labels = label_image(
            retrianed_graph_textbox.value,
            test_image_path_text_box.value,
            labels_textbox.value,
        )
        inference_result.value = "<b>" + labels[0] + "<b>"
        inference_result.description = "result: "

    image_inference_button.on_click(image_inference_button_clicked)

    return retrain_images_box



from mammut.curriculum.report.messages.message import CurriculumBaseMessage
from mammut.curriculum.report.messages.curriculum_messages import *
from typing import List
import ray
import logging

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


@ray.remote
class RayReportersRegistry:
    """The RayReporterRegistry actor is used in remote workers as the actor
    handle to centralize all the remote produced messages.

    The main curriculum then can request messages in a single place
    and report them where the user "is".
    """

    def __init__(self):
        self._cached_messages: List[CurriculumBaseMessage] = []
        self._last_index_retrieved: int = 0

    def get_messages(self):
        """Returns the new messages arrived since the last
        call to this method.
        """
        messages = self._cached_messages[
            self._last_index_retrieved : len(self._cached_messages)
        ]
        self._last_index_retrieved = len(self._cached_messages)
        log.debug(f"Retrieve {len(messages)} new messages")
        return messages

    def save_message(self, message: CurriculumBaseMessage):
        """Adds a new message to the report registry.
        """
        self._cached_messages.append(message)

    def save_ktt_plain_message(self, message: str, level: str):
        """Receives a plain string message and the plain
        report string level.
        Internally, the string message will be reported as some
        curriculum message type.
        Note: This method is used for allowing mammut-ktt to publish
        messages in the report registry without dealing with serialization
        code of more complex classes like messages.
        Args:
            message: the plain string message
            level: the report level plain string.
        Returns
            Returns an int, without especial meaning. Just to handshake the
            Java call in the mammut-ktt client code.
        """
        log_level = logging.getLevelName(level)
        if log_level == logging.DEBUG:
            report_message = CurriculumGeneralDebugMessage(message)
        elif log_level == logging.INFO:
            report_message = CurriculumGeneralInfoMessage(message)
        elif log_level == logging.ERROR:
            report_message = CurriculumGeneralErrorMessage(message)
        elif log_level == logging.WARNING:
            report_message = CurriculumGeneralWarningMessage(message)
        else:
            report_message = CurriculumGeneralInfoMessage(message)
        self._cached_messages.append(report_message)
        log.debug("received KtT plain message")
        return 0

    def get_all_messages(self):
        """Returns all the messages arrived since this process
        started.
        """
        return self._cached_messages



from enum import Enum


class CorpusTable(Enum):
    """
    Enum with the list of constant required by the Corpus table.
    """

    # Column names
    ID_COLUMN_NAME = "id"
    SUB_ID_COLUMN_NAME = "sub_id"
    SCENARIO_TYPE_COLUMN_NAME = "scenario_type"
    EVENT_MESSAGE_COLUMN_NAME = "event_message"
    HIDDEN_COLUMN_NAME = "hidden"
    FIELD_COLUMN_NAME = "field"
    LAMBDA_CONDITION_COLUMN_NAME = "lambda_condition"
    UI_EVENT_COLUMN_NAME = "ui_event"
    ACTION_COLUMN_NAME = "action"
    SOURCE_COLUMN_NAME = "source"
    REGIONAL_SETTINGS_COLUMN_NAME = "regional_settings"
    COMPLEXITY_COLUMN_NAME = "complexity"
    ISSUE_TYPE_COLUMN_NAME = "issue_type"
    DATE_COLUMN_NAME = "date"
    # Other constants
    MAMMUT_USERS_PREFIX = "mammut"

    def __repr__(self):
        return str(self.value)

    def __call__(self, *args, **kwargs):
        return str(self.value)


class CrawlerCorpusTable:
    """Data class to encapsulate Crawler Corpus default
    constant values.

    Defined here to avoid circular dependencies.
    """
    DEFAULT_COLUMN_TITTLES = [
        [
            CorpusTable.ID_COLUMN_NAME(),
            CorpusTable.SUB_ID_COLUMN_NAME(),
            CorpusTable.SCENARIO_TYPE_COLUMN_NAME(),
            CorpusTable.EVENT_MESSAGE_COLUMN_NAME(),
            CorpusTable.HIDDEN_COLUMN_NAME(),
            CorpusTable.FIELD_COLUMN_NAME(),
            CorpusTable.SOURCE_COLUMN_NAME(),
            CorpusTable.REGIONAL_SETTINGS_COLUMN_NAME(),
            CorpusTable.COMPLEXITY_COLUMN_NAME(),
        ]
    ]
    DEFAULT_COLUMN_TITTLE_RANGE = "!A2"
    DEFAULT_START_COLUMN_TITTLE = "A"
    DEFAULT_START_ROW_TITTLE = "2"
    URL_KEY = "url"
    NODES_PATTERN_KEY = "nodes_pattern"
    NODE_ELEMENT_KEY = "node_elements"
    PAGE_HREF_KEY = "page_href"
    DEFAULT_PAGES_LIMIT_KEY = "pages_limit"
    PAGINATION_NODE_PATTERN_KEY = "pagination_node_pattern"
    PAGINATION_NODE_ELEMENT_KEY = "pagination_node_element"
    DEFAULT_SCENARIO_TYPE_VALUE = "Monologue"
    DEFAULT_EVENT_MESSAGE_VALUE = ""
    DEFAULT_HIDDEN_VALUE = ""
    DEFAULT_FIELD_VALUE = ""
    DEFAULT_SOURCE_VALUE = "CrawlerCorpus"
    DEFAULT_COMPLEXITY_VALUE = ""
    DEFAULT_CRAWLING_PERIOD = 86400  # 1 day






from abc import abstractmethod, ABC
from mammut.curriculum.report.messages.message import CurriculumBaseMessage
import logging

log = logging.getLogger(__name__)


class BaseReporter(ABC):
    """Generic interface for Report components.

    A Reporter will report system messages that are important
    to users. These Messages uses the level threshold
    provided in python 'logging' package.

    Reporters use the same threshold level than python logging package.
    """

    def __init__(self, threshold=logging.INFO):
        """
        Args:
            threshold: threshold for this reporter.
        """
        self.threshold = threshold

    @abstractmethod
    def send_message(self, message: CurriculumBaseMessage, **kwargs):
        """
        Args:
            message: message to be reported.
            kwargs: any other required concrete parameter.
        """
        pass



from typing import Optional
from mammut.curriculum.models.model_base import ModelBase
from transformers import (
    pipeline,
    AutoTokenizer,
    AutoModelForSequenceClassification,
    TFAutoModelForSequenceClassification,
)
from abc import abstractmethod
import torch
import tensorflow as tf
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class HuggingFaceSequenceClassificationModel(ModelBase):
    """Downloads and caches a HuggingFace pretrained model and its tokenizer.

    This is a base API for 'sequence classification' in Hugging Face models.
    Sequence classification is the task of classifying sequences according to a
    given number of classes.

    ..  _Hugging Face Docs Sequence Classification:
        https://huggingface.co/transformers/task_summary.html#sequence-classification
    """

    def __init__(self):
        ModelBase.__init__(self)
        self.pt_model: Optional[AutoModelForSequenceClassification] = None
        self.tf_model: Optional[TFAutoModelForSequenceClassification] = None
        self.tokenizer: Optional[AutoTokenizer] = None

    @property
    @abstractmethod
    def hf_model_name(self) -> str:
        """Returns a Hugging Face's model name.

        This method is used to provide main class a Hugging Face's model name
        to be downloaded and cached locally.

        ..  _Hugging Face Docs summary of the models:
            https://huggingface.co/transformers/model_summary.html
        """
        pass

    @property
    @abstractmethod
    def task(self) -> str:
        """Returns a task name for Hugging Face models.

        Hugging Face models allow for Natural Language Understanding (NLU) tasks
        such as question answering, sequence classification, named entity
        recognition and others.

        ..  _Hugging Face Docs summary of the tasks:
            https://huggingface.co/transformers/task_summary.html
        """
        pass

    @abstractmethod
    def paraphrase_identification(self, input_1: str, input_2: str) -> int:
        """ It performs a comparison between two inputs to determine
        how close they are to be a paraphrase one another.
        """
        pass

    def sentiment_analysis(self, text_input):
        """Hugging Face's pipeline task.

        Args:
            text_input(str): input text to perform analysis.

        Return:
            Returns a label alongside a score.
        """
        results = self.classifier(text_input)
        return results

    def train(self, mammut_session_context: MammutSessionContext, **kwargs):
        """Passes ModelBase's train method.

        Train method passes since class uses a pretrained model.
        """
        pass

    def save(self, mammut_session_context: MammutSessionContext, **kwargs):
        """Models persistence hasn't been implemented"""
        pass


class TorchHuggingFaceSequenceClassification(
    HuggingFaceSequenceClassificationModel
):
    """Allows to perform text classification tasks in Hugging Face models using PyTorch Framework.

    Text classification is the task of classifying text according to
    different criteria. An example can be comparing two inputs to see
    if they're paraphrases one another.
    """

    def __init__(self):
        HuggingFaceSequenceClassificationModel.__init__(self)
        self.pt_model = AutoModelForSequenceClassification.from_pretrained(
            self.hf_model_name
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.hf_model_name)
        self.classifier = pipeline(
            task=self.task, model=self.pt_model, tokenizer=self.tokenizer
        )

    def hf_model_name(self) -> str:
        pass

    @property
    def task(self) -> str:
        """Provides a Hugging Face's pipeline task name.

        Returns:
            Returns task name.
        """
        return "sentiment-analysis"

    def paraphrase_identification(self, input_1: str, input_2: str) -> int:
        """It performs a comparison between two inputs to determine
        how close they are to be a paraphrase one another.

        Args:
            input_1(str): first input text.
            input_2(str): second input text.

        Returns:
            Returns a classification score.
        """
        paraphrase = self.tokenizer(input_1, input_2, return_tensors="pt")
        paraphrase_classification_logits = self.pt_model(**paraphrase).logits
        paraphrase_results = torch.softmax(
            paraphrase_classification_logits, dim=1
        ).tolist()[0][1]
        return paraphrase_results


class TensorFlowHuggingFaceSequenceClassification(
    HuggingFaceSequenceClassificationModel
):
    """Allows to perform text classification tasks in Hugging Face models using TensorFlow Framework.

    Text classification is the task of classifying text according to
    different criteria. An example can be comparing two inputs to see
    if they're paraphrases one another.
    """

    def __init__(self):
        HuggingFaceSequenceClassificationModel.__init__(self)
        self.tf_model = TFAutoModelForSequenceClassification.from_pretrained(
            self.hf_model_name, from_pt=True
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.hf_model_name)
        self.classifier = pipeline(
            task=self.task, model=self.pt_model, tokenizer=self.tokenizer
        )

    def hf_model_name(self) -> str:
        pass

    @property
    def task(self) -> str:
        """Provides a Hugging Face's pipeline task name.

        Returns:
            Returns task name.
        """
        return "sentiment-analysis"

    def paraphrase_identification(self, input_1: str, input_2: str) -> int:
        """It performs a comparison between two inputs to determine
        how close they are to be a paraphrase one another.

        Args:
            input_1(str): first input text.
            input_2(str): second input text.

        Returns:
            Returns a classification score.
        """
        paraphrase = self.tokenizer(input_1, input_2, return_tensors="tf")
        paraphrase_classification_logits = self.tf_model(paraphrase).logits
        paraphrase_results = tf.nn.softmax(
            paraphrase_classification_logits, axis=1
        ).numpy()[0][1]
        return paraphrase_results


class TorchRuperta(TorchHuggingFaceSequenceClassification):
    """Hugging Face model 'Ruperta' integration.
    """

    def __init__(self):
        TorchHuggingFaceSequenceClassification.__init__(self)

    @property
    def hf_model_name(self) -> str:
        """Provides a Hugging Face's model name.

        Returns:
            Returns model name.

        ..  _Hugging Face RuPERTa-base-finetuned-pawsx-es:
            https://huggingface.co/mrm8488/RuPERTa-base-finetuned-pawsx-es
        """
        return "mrm8488/RuPERTa-base-finetuned-pawsx-es"


class TensorFlowRuperta(TensorFlowHuggingFaceSequenceClassification):
    """Hugging Face model 'Ruperta' integration.

    RuPERTa is a paraphrase identification model.

    ..  _Hugging Face RuPERTa-base-finetuned-pawsx-es:
        https://huggingface.co/mrm8488/RuPERTa-base-finetuned-pawsx-es
    """

    def __init__(self):
        TensorFlowHuggingFaceSequenceClassification.__init__(self)

    @property
    def hf_model_name(self) -> str:
        """Provides a Hugging Face's model name.

        Returns:
            Returns model name.
        """
        return "mrm8488/RuPERTa-base-finetuned-pawsx-es"



import hashlib

from mammut.common.basic_tokenizer import BasicTokenizer
from mammut.common.corpus.constants import CorpusTable
from mammut.common.corpus.corpus_base import ScenarioType
from mammut.common.corpus.natural_corpus import NaturalCorpus
from mammut.common.lexicon.dictionary import Dictionary


class DictionaryCorpus(NaturalCorpus):
    DEFAULT_SHEET_NAME = "dictionary_corpus"

    def __init__(
        self,
        spreadsheet_id: str,
        id,
        corpus_sheet_title: str,
        corpus_extension_sheet_title: str,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        regional_settings: str,
        annotable: bool,
        hash_lemma_dict: dict,
        corpus_dataframe,
    ):
        NaturalCorpus.__init__(
            self,
            spreadsheet_id,
            id,
            corpus_sheet_title,
            corpus_extension_sheet_title,
            tokenizer,
            dictionary,
            regional_settings,
            annotable,
            corpus_dataframe,
        )
        self.hash_lemma_dict = hash_lemma_dict

    @staticmethod
    def lemma_entry_to_events_rows(hashed_lemma, lemma_entry):
        events_list = list()
        for pos in lemma_entry.postags_entries:
            for definition_set in pos.definitions_set_entries:
                for definition_entry in definition_set.definitions_entries:
                    event_message = list()
                    event_message.append(hashed_lemma)  # id
                    event_message.append(
                        definition_set.definition_index
                    )  # sub_id
                    event_message.append(
                        ScenarioType.Monologue.value
                    )  # scenario_type
                    event_message.append(
                        definition_entry.definition
                    )  # event_message
                    event_message.append("")  # hidden
                    event_message.append("")  # field
                    event_message.append("dictionary")  # source
                    event_message.append(
                        definition_entry.regional_settings_definition
                    )  # regional seting
                    event_message.append("")  # complexity
                    events_list.append(event_message)
        return events_list

    @classmethod
    def transform_dictinary_to_sheet_rows(cls, dictionary):
        df_columuns = [
            CorpusTable.ID_COLUMN_NAME(),
            CorpusTable.SUB_ID_COLUMN_NAME(),
            CorpusTable.SCENARIO_TYPE_COLUMN_NAME(),
            CorpusTable.EVENT_MESSAGE_COLUMN_NAME(),
            CorpusTable.HIDDEN_COLUMN_NAME(),
            CorpusTable.FIELD_COLUMN_NAME(),
            CorpusTable.SOURCE_COLUMN_NAME(),
            CorpusTable.REGIONAL_SETTINGS_COLUMN_NAME(),
            CorpusTable.COMPLEXITY_COLUMN_NAME(),
        ]
        hash_lemma_dict = dict()
        sheet_rows = list()
        sheet_rows.append(df_columuns)
        for scenario_index, lemma_entry in enumerate(
            dictionary.lemmas_entries
        ):
            hashed_lemma = int(
                hashlib.sha256(lemma_entry.lemma.encode("utf-8")).hexdigest(),
                16,
            ) % (10 ** 8)
            hash_lemma_dict[hashed_lemma] = lemma_entry.lemma
            lemma_rows = cls.lemma_entry_to_events_rows(
                hashed_lemma, lemma_entry
            )
            sheet_rows = sheet_rows + lemma_rows
        return sheet_rows, hash_lemma_dict



from mammut.common.corpus.corpus_operations import CorpusOperations
from typing import List, Dict
from abc import abstractmethod
import datasets


class HuggingFaceDatasetCorpus(CorpusOperations):
    """Generic interface for any type of curriculum training corpus
    based on a Hugging Face datasets.Dataset.

    Any concrete HuggingFaceDatasetCorpus known how to transform itself
    to a datasets.Dataset. By doing so, it straightforward to use it as
    training dataset of a Hugging Face model wrapper.

    This type of corpus doesn't form part of Mammut Corpus class hierarchy.
    However, It would be possible to create a Mammut Corpus class that implements
    the transformation to a datasets.Dataset. If so, this interface should be used.

    Currently, this kind of corpus isn't indexable or annotable.
    """

    def __init__(self, package_spreadhseet_id: str, corpus_id: str):
        """
        Args:
            package_spreadhseet_id:
                Package spreadsheet ID.
            corpus_id:
                ID assigned to this corpus in the CorpusMap column.
        """
        self.corpus_id = corpus_id
        self.spreadsheet_id = package_spreadhseet_id

    def get_events_as_strings(self) -> List[str]:
        raise NotImplementedError

    @property
    def indexable(self) -> bool:
        """This Corpus isn't considered for mammut-py indexing process."""
        return False

    @property
    def annotable(self) -> bool:
        """This Corpus isn't considered for mammut-py annotation process."""
        return False

    @abstractmethod
    def get_hf_dataset(self) -> datasets.Dataset:
        """Return a datasets.Dataset that can be used as training data for
        a hugging face model integration.
        """
        pass


class HuggingFaceFixedCorpus(HuggingFaceDatasetCorpus):
    """HuggingFace corpus for any fixed dataset provided as it is in
    datasets hub.

    The term fixed implies that the get_hf_dataset will return the
    downloaded dataset preserving the dataset.DatasetInfo configuration.

    If there's a requirement for changing the dataset.DatasetInfo, try to
    specify those operations in:
        - A specific child class of HuggingFaceDatasetCorpus. (Recommended).
        - In the curriculum Lesson/Model class that is intended to use the corpus.
    """
    def __init__(
        self, package_spreadsheet_id: str, corpus_id: str, config_kwargs: Dict
    ):
        """
        Args:
            package_spreadsheet_id:
                Package spreadsheet ID.
            corpus_id:
                ID assigned to this corpus in the CorpusMap column.
            config_kwargs:
                Dictionary to forward as the datasets.load_dataset **config_kwargs.

        """
        super(HuggingFaceFixedCorpus, self).__init__(
            package_spreadsheet_id, corpus_id
        )

        if "dataset_build_args" not in config_kwargs:
            raise RuntimeError(f"HuggingFace fixed corpus without dataset_build_args. CorpusID: {corpus_id}")

        build_args = config_kwargs["dataset_build_args"]
        self._base_dataset = datasets.load_dataset(**build_args)

        if "rename_columns" in config_kwargs:
            columns_rename_dict: Dict[str, str] = config_kwargs["rename_columns"]
            self._rename_dataset_columns(columns_rename_dict)

    def get_hf_dataset(self) -> datasets.Dataset:
        return self._base_dataset

    def _rename_dataset_columns(self, columns_rename_dict: Dict[str, str]):
        """Renames the columns in the self._dataset.

        Args:
            columns_rename_dict: A dictionary of Keys as the current column name,
                and the corresponding Values as the desired column name.

        Raises:
            ValueError if some Key isn't found in the dataset columns.

        Returns:
            None. This method renames the columns as a side effect.
        """
        for column_old_name, column_new_name in columns_rename_dict.items():
            dataset_columns = self._base_dataset.column_names
            if column_old_name not in dataset_columns:
                raise ValueError(f'HuggingFace Fixed corpus rename error. Column "{column_old_name}" not found in dataset.')

            self._base_dataset.rename_column_(column_old_name, column_new_name)



from mammut.curriculum.lesson import lesson_type
import mammut.curriculum.core.utils as curriculum_utils
from mammut.curriculum.lesson import lesson_base
from mammut.curriculum.lesson.online_lessons import KttInterpretationLesson
from mammut.assessment.assessor import Assessor, Assessment
from typing import List, Optional
import pandas as pd
from mammut.common.corpus.corpus_map import CorpusMap
from mammut.curriculum.course.course_state import CourseState
from mammut.curriculum.core.mammut_session_context import MammutSessionContext
import mammut.curriculum.report as report


class Course(CourseState):
    """This class encapsulates the concept of a course in the curriculum.

    A course is composed essentially by a ordered list of lessons.
    Each Lesson has a set of scheduled assessments, composed of problems.
    Problems encapsulate a weight, which is used to compute the
    course score for approval/disapproval.

    The computation role of the course involves:
    - Prepare a lesson resources: input data, model.
    - Practice a Lesson to train the model.
    - After a Lesson is finished, ask for scheduled assessments
        to that lesson.
    - Apply scheduled assessments for that finished lesson.
    - Repeat for each lesson.
    """

    """
    Registry for lessons type. This internal registry allows to 
    instantiate concrete lessons type using a dictionary key lookup.
    """
    _lessons_registry = {}

    """
    Loads the lessons type internal dictionary used for
    concrete instantiation of lessons.

    The dictionary keys are enumeration names for the problem type, 
    and the values are problems custom classes.
    """
    for enum_type in lesson_type.LessonType.list():
        _lessons_registry[enum_type[0]] = enum_type[1]

    def __init__(
        self,
        course_id: str,
        academic_period: str,
        prerequisites: str,
        language_level: str,
        course_name: str,
        approval_criteria: str,
        retry: str,
        regional_settings: str,
        lessons: List[lesson_base.Lesson],
        assessor: Assessor,
        mammut_session_context: MammutSessionContext,
    ):
        """
        Note: some arguments are received as string values from storage manager retrieved
        dataset.

        Args:
            course_id(str): 'id' value in datasource retrieved data.
            academic_period(str): 'academic_period' value in datasource retrieved data.
            prerequisites(str): 'prerequisites' value in datasource retrieved data.
            language_level(str): 'language_level' value in datasource retrieved data.
            course_name(str): 'course_name' value in datasource retrieved data.
            retry(str): 'retry' string value to indicate how many times a course can be processed before
                stopping the curriculum.
            regional_settings(str): 'regional_settings' value in datasource retrieved data.
            lessons: Lessons for this course
            assessor: assessor reference for course assessment operations
            mammut_session_context: The session context from the classroom.
        """

        # Splits comma separated list of prerequisites for a course.
        prerequisites_list: List[int] = (
            []
            if not prerequisites
            else [
                int(prerequisite.strip())
                for prerequisite in prerequisites.split(",")
            ]
        )

        language_level: curriculum_utils.LanguageLevel = curriculum_utils.LanguageLevel.from_str(
            language_level
        )

        """
        Lessons list for the course.
        """
        self.lessons: List[lesson_base.Lesson] = lessons

        """
        Assessor reference should be assigned by the subclass.
        """
        self.assessor: Assessor = assessor

        perfect_score = self.assessor.get_course_max_score(int(course_id))

        self.course_id = int(course_id)
        self.academic_period = int(academic_period)
        self.prerequisites = prerequisites_list
        self.language_level = language_level
        self.course_name = course_name
        self._approval_criteria = float(approval_criteria)
        self._retry_value = int(retry)
        self._retries_count = 0
        self.regional_settings = regional_settings
        self._last_score = 0.0
        self._perfect_score = perfect_score
        self._mammut_session_context = mammut_session_context

    @property
    def last_attempt(self) -> int:
        """Returns the current number of processed attempts for the course"""
        return self._retries_count

    @property
    def last_score(self) -> float:
        return self._last_score

    def is_allowed(self) -> bool:
        """Returns true if the course is allowed for processing.

        A course might not be allowed for processing if:
            - Number of retries has been reached.

        More conditions can be added in the future.
        """
        retry_allowed = self._retries_count <= self._retry_value
        return retry_allowed

    def start(
        self,
        progress_writer: "Progress",
        mammut_session_context: MammutSessionContext,
    ):
        """Start processing internal lessons in this course.

        Lessons are an ordered list. They are executed one after other.
        Restart _last_score for a new "start" called.

        Args:
            progress_writer: object to write progress in the package document.
            mammut_session_context: session context information for report.
        """
        self._last_score = 0.0
        self._retries_count += 1
        for le in self.lessons:
            """
            Lectures all the prepared knowledge for the lesson.
            """
            report.send_message(
                report.CurriculumGeneralInfoMessageWithLesson(
                    "Lesson lecturing", self.course_id, le.lesson_id
                )
            )
            le.lecture()

            """
            After the lesson has been prepared, it can be practiced.
            """
            report.send_message(
                report.CurriculumGeneralInfoMessageWithLesson(
                    "Lesson practice", self.course_id, le.lesson_id
                ),
            )
            le.practice(mammut_session_context)

            """
            Request scheduled assessment for practiced lesson
            Apply assessment to practiced lesson, and 
            adds the approved points to the _last_score
            """
            scheduled_assessment = self._request_assessments(le.order)
            if scheduled_assessment is not None:
                self._last_score += scheduled_assessment.apply(le)
                progress_writer.add_progress_entry(self, scheduled_assessment)

    def approved(self) -> bool:
        """Returns true if the Course has been approved.

        The course is approved if all the applied assessments
        were approved.
        """
        return (
            self._last_score / self._perfect_score
        ) >= self._approval_criteria

    def _request_assessments(
        self, finished_lesson_order: int
    ) -> Optional[Assessment]:
        """Returns a list of scheduled assessments for the finished lesson.
        """
        return self.assessor.get_scheduled_assessment(
            self.course_id, finished_lesson_order
        )

    @staticmethod
    def get_lessons_for_course(
        course_id: str,
        lessons_df: pd.DataFrame,
        corpus_map: CorpusMap,
        mammut_session_context: MammutSessionContext,
    ) -> List[lesson_base.Lesson]:
        lessons: List[lesson_base.Lesson] = []
        for index, row_lesson in lessons_df.loc[
            lessons_df[lesson_base.Lesson.COURSE_COLUMN] == course_id
        ].iterrows():
            """
            Creates the Lesson based on the returned class
            reference 
            """
            lesson = Course._lessons_registry[
                row_lesson[lesson_base.Lesson.MODEL_COLUMN]
            ](
                row_lesson[lesson_base.Lesson.ID_COLUMN],
                row_lesson[lesson_base.Lesson.COURSE_COLUMN],
                row_lesson[lesson_base.Lesson.ORDER_COLUMN],
                row_lesson[lesson_base.Lesson.CORPUS_ID_COLUMN],
                corpus_map,
                row_lesson[lesson_base.Lesson.NAME_COLUMN],
                row_lesson[lesson_base.Lesson.LANGUAGE_COMPETENCE_COLUMN],
                row_lesson[lesson_base.Lesson.PARAMETERS_COLUMN],
                mammut_session_context,
            )
            lessons.append(lesson)

        lessons.sort(key=lambda x: x.order)
        return lessons

    def __hash__(self):
        return self.course_id

    def __eq__(self, other):
        return (
            isinstance(other, CourseState)
            and self.course_id == other.course_id
        )

    def get_course_record(self) -> CourseState:
        return CourseState(
            self.course_id,
            self.academic_period,
            self.prerequisites,
            self.language_level,
            self.course_name,
            self._approval_criteria,
            self._retry_value,
            self._retries_count,
            self.regional_settings,
            self._last_score,
            self._perfect_score,
            self._mammut_session_context,
        )

    def fetch_course_state(self, record: CourseState):
        """Updates the course state to the received course state.

        It will keep the greater retries count.
        """
        if not self.course_id == record.course_id:
            raise ValueError(
                "Can't updated a course state with different course_id"
            )

        if record._retries_count > self._retries_count:
            self._retries_count = record._retries_count
        self._last_score = record._last_score


class KttPackageInterpretationCourse(Course):
    """This course is the responsible for handling the Package interpretation
    process.

    The package compilation process is a pre-requisite to all courses,
    so if enabled, it's executed in the academic period "0".

    KtT transducer will perform the interpretation and store the data in the
    long term working memory.

    Todo: Evaluation and assessment process for this course
        isn't defined yet.

    Todo: synchronization between classroom and remote java process
        isn't implemented yet. Some wait should be useful.

    Todo: Language level isn't defined yet for this course. Using
        `SemanticPragmatic` as some value.

    Todo: Remember to implement the mammut-id retrieval when mammut
        entity provider generalization are implemented.
        (https://github.com/mammut-io/mammut-py/issues/150)
    """

    def __init__(
        self, mammut_session_context: MammutSessionContext, assessor: Assessor
    ):
        super(KttPackageInterpretationCourse, self).__init__(
            course_id="0",
            academic_period="0",
            prerequisites="",
            language_level="SemanticPragmatic",
            course_name="KtT package interpretation",
            approval_criteria="1.0",
            retry="1",
            regional_settings="en",
            lessons=[],
            assessor=assessor,
            mammut_session_context=mammut_session_context,
        )

        self._perfect_score = 1.0

        ktt_interpretation_lesson = KttInterpretationLesson(
            lesson_id=KttInterpretationLesson.RESERVED_LESSON_ID,
            course=f'{self.course_id}',
            order="1",
            corpus_id="0",
            name=self.course_name,
            language_competence="WordIdentification",
            parameters="{}",
            mammut_session_context=mammut_session_context,
        )

        self.lessons.append(ktt_interpretation_lesson)

    def start(
        self,
        progress_writer: "Progress",
        mammut_session_context: MammutSessionContext,
    ):
        super(KttPackageInterpretationCourse, self).start(
            progress_writer, mammut_session_context
        )
        # This course is approved after package compilation has been performed.
        self._last_score = self._perfect_score



from mammut.curriculum.report.errors.base_error import BaseError
from mammut.curriculum.report.utils import Phase
from json import JSONDecodeError


class JSONLessonParsingError(BaseError):
    def __init__(
        self,
        exception: JSONDecodeError,
        phase: Phase,
        course: str,
        lesson: str,
        column: str,
    ):
        BaseError.__init__(self, "Error parsing JSON string for Lesson", phase)
        self.course: str = course
        self.lesson: str = lesson
        self.column: str = column
        self.exception = exception

    def __str__(self):
        s = (
            f"{self.message}, "
            f"Phase: {self.phase}, "
            f"Course: {self.course}, "
            f"Lesson: {self.lesson}, "
            f"Column value: {self.column}"
        )
        return s


class JSONProblemParsingError(BaseError):
    def __init__(
        self,
        exception: JSONDecodeError,
        phase: Phase,
        course_id: str,
        assessment_id: str,
        problem_id: str,
        column_value: str,
    ):
        BaseError.__init__(
            self, f"Error parsing JSON string for Problem", phase
        )
        self.phase = phase
        self.course = course_id
        self.assessment = assessment_id
        self.problem = problem_id
        self.column = column_value
        self.exception = exception

    def __str__(self):
        s = (
            f"{self.message}, "
            f"Phase: {self.phase}, "
            f"Course: {self.course}, "
            f"Assessment: {self.assessment}, "
            f"Problem:{self.problem}, "
            f"Column: {self.column}"
        )
        return s


class ModelRequiredParameterMissingKey(BaseError):
    def __init__(self, missing_key: str, course_id, lesson_id):
        BaseError.__init__(
            self, f'"Missing key {missing_key} in lesson {lesson_id} from course {course_id}"', Phase.LECTURE_LESSON
        )

        self.course_id = course_id
        self.lesson_id = lesson_id

    def __str__(self):
        return self.message



# coding=utf-8
import logging
import mammut.curriculum.core.utils as curriculum_utils
from mammut.common.corpus.corpus_map import CorpusMap
from abc import ABC, abstractmethod
from typing import List
import enum
from typing import Optional, Dict
import json
from mammut.curriculum.models.model_base import ModelBase
from mammut.curriculum.report.errors.json_parsing_errors import (
    JSONLessonParsingError,
)
from mammut.curriculum.report.utils import Phase
from mammut.curriculum.core.mammut_session_context import MammutSessionContext

log = logging.getLogger(__name__)


class CorpusType(enum.Enum):
    """Enumeration type for corpus supported in Lessons.

    Currently supported:
        - synthetic
        - raw
    """

    Synthetic = 1
    Raw = 2

    @staticmethod
    def from_str(label):
        if label == "synthetic":
            return CorpusType.Synthetic
        elif label == "raw":
            return CorpusType.Raw
        else:
            raise NotImplementedError


class Lesson(ABC):
    """Lesson class models a unit of model training
    and evaluation.

    A Lessons encapsulates:
    - A model to train
    - Parameters for the model
    - Input data

    A lesson should be subclassed based on the model to be used.
    Each lesson is associated with a course.
    """

    ID_COLUMN = "id"
    COURSE_COLUMN = "course"
    ORDER_COLUMN = "order"
    CORPUS_ID_COLUMN = "corpus_id"
    MODEL_COLUMN = "model"
    NAME_COLUMN = "name"
    LANGUAGE_COMPETENCE_COLUMN = "language_competence"
    PARAMETERS_COLUMN = "parameters"

    DEFAULT_START_CELL = "A1"
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "1"
    DEFAULT_END_CELL = "H"

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        corpus_map: Optional[CorpusMap],
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        """
        Note: Arguments are received as strings from storage manager retrieved
        dataset.

        Args:
            lesson_id(str): 'id' value in datasource retrieved data.
            course(str): 'course' value in datasource retrieved data.
            order(str): 'order' value in datasource retrieved data.
            corpus_id(str): 'corpus_id' value in datasource retrieved data.
            corpus_map: reference to the Package corpus map object.
            name(str): 'name' value in datasource retrieved data.
            language_competence(str): 'language_competence' value in datasource retrieved data.
            parameters(str): json string parameters for the model.
        """
        self.lesson_id: int = int(lesson_id)
        self.course: int = int(course)
        self.order: int = int(order)
        self.corpus_id: int = int(corpus_id)
        """
            Currently, only supported plain corpus as input data. 
            (A list of string tokens). More elaborated corpus elements 
            might be necessary in the future. 
        """
        self.name = name
        self.language_competence: curriculum_utils.LanguageCompetence = curriculum_utils.LanguageCompetence.from_str(
            language_competence
        )
        try:
            self.parameters_dict: Dict = json.loads(
                parameters.replace("\n", "")
            )
        except json.JSONDecodeError as e:
            raise JSONLessonParsingError(
                e,
                Phase.LOAD_CURRICULUM_LESSONS_DATA,
                course,
                lesson_id,
                self.PARAMETERS_COLUMN,
            )
        self._mammut_session_context: MammutSessionContext = mammut_session_context
        self._corpus_map_reference: CorpusMap = corpus_map
        self._model: ModelBase = self._get_model_instance()
        self.corpus = Optional[None]

    @abstractmethod
    def _get_model_instance(self) -> ModelBase:
        """Factory method for concrete instantiation of model at
        initialization time
        :param **kwargs: """
        pass

    @abstractmethod
    def lecture(self):
        """Loads all the knowledge resources required to train the model.

        This includes (but is not limited to) for example:
        - Load the training data for the model
        - Load pre-trained models if required
        - Pre-train a part of the model
        """
        pass

    @abstractmethod
    def practice(self, mammut_session_context: MammutSessionContext):
        """Process the training of the model and save the model after
        training.

        Note: each Lesson subclass can implement specific code to
        this method according to the specific language model object
        it's using.

        Args:
            save_path: file system folder path to save the model.
        """
        pass

    @property
    def model(self) -> ModelBase:
        """Model used by the lesson.

        Models are integrated through the curriculum.models.base.ModelBase
        interface.

        Returns:
            The object instance of the model,
                self._model.
        """
        return self._model

    @staticmethod
    def get_corpus(corpus_id: int, corpus_map: CorpusMap,) -> List[str]:
        """Returns a training corpus from the corpus map.

        Args:
            corpus_id(int): corpus id in corpus map sheet
                in package.
            corpus_map: CorpusMap object from within the
                Package used.

        Returns:
            The corpus instance.
        """
        corpus = next(
            filter(
                lambda c: c.id == str(corpus_id), corpus_map.get_all_corpus()
            ),
            None,
        )
        if corpus is None:
            log.info(f"Training Corpus not found. Corpus ID: {corpus_id}")
            raise RuntimeError(
                f"Training Corpus not found. Corpus ID: {corpus_id}"
            )
        return corpus.get_events_as_strings()



# coding=utf-8
from mammut.common.mammut_api import *
from mammut.common.simulator_utilities import *
from ipywidgets import *
from datetime import datetime

### User Layout ###
def get_user_form():
    form_values = dict()
    personUserDropdown = widgets.Dropdown(
        description="Usuario:", disabled=False, button_style=""
    )
    form_values["personUserDropdown"] = personUserDropdown
    supervisorUserDropdown = widgets.Dropdown(
        description="Supervisor:", disabled=False, button_style=""
    )
    form_values["supervisorUserDropdown"] = supervisorUserDropdown

    languages = {}
    languages["Español"] = "es"
    languages["Ingles"] = "en"

    newMammutLayout = Layout(width="100%", height="50px")
    form_values["newMammutLayout"] = newMammutLayout
    nameTextBox = widgets.Text(
        value="",
        placeholder="Mammut Restaurant",
        description="Nombre:",
        layout=newMammutLayout,
    )
    form_values["nameTextBox"] = nameTextBox
    emailTextBox = widgets.Text(
        value="",
        placeholder="restaurant@mammut.io",
        description="Email:",
        layout=newMammutLayout,
    )
    form_values["emailTextBox"] = emailTextBox
    birthDateBox = widgets.Text(
        value="",
        placeholder="dd/mm/aaaa",
        description="Fecha de nacimiento:",
        layout=newMammutLayout,
    )
    form_values["birthDateBox"] = birthDateBox
    phoneTextBox = widgets.Text(
        value="",
        placeholder="+582129769097",
        description="Telefono:",
        layout=newMammutLayout,
    )
    form_values["phoneTextBox"] = phoneTextBox
    facebookIdTextBox = widgets.Text(
        value="",
        placeholder="301598836881228",
        description="Facebook Id:",
        layout=newMammutLayout,
    )
    form_values["facebookIdTextBox"] = facebookIdTextBox
    fbAccessTokenTextBox = widgets.Text(
        value="",
        placeholder="EAAZAbcJHTWr4BACnPewhF08OGZBZBZCOKyBn3rK9PDFveMJ4s1xHo0JXdXNgpKXveYOd3hj0spLEwJk0cAtt2Dt6Wl4ukaTjoRQCMvOS0hxFUSbNJp5T738VrCPN7q8XWp3QZBreGhyj8nEeLsZAhfxuJ9vZB9fGyl2kkpHPtzHiAZDZD",
        description="FB Page Access Token:",
        layout=newMammutLayout,
    )
    form_values["fbAccessTokenTextBox"] = fbAccessTokenTextBox
    langDropdown = widgets.Dropdown(
        options=languages, value="es", description="Idioma:"
    )
    form_values["langDropdown"] = langDropdown
    user_types = {user_type.value: user_type for user_type in UserType}
    userType = widgets.Dropdown(
        options=user_types, value=UserType.person, description="User Type:"
    )
    form_values["userType"] = userType
    lastname_style = {"description_width": "initial"}
    lastNameTextBox = widgets.Text(
        value="",
        placeholder="appellido",
        description="Apellido:",
        style=lastname_style,
        layout=newMammutLayout,
    )
    form_values["lastNameTextBox"] = lastNameTextBox

    newMammutButton = widgets.Button(
        description="Crear Nuevo ...", button_style="primary"
    )

    userMammutLayout = Layout(width="50%", height="80px", visibility="visible")

    userBox = widgets.VBox(
        [
            widgets.HBox(
                [
                    widgets.VBox(
                        [
                            nameTextBox,
                            lastNameTextBox,
                            emailTextBox,
                            birthDateBox,
                            langDropdown,
                        ],
                        layout=Layout(width="100%"),
                    ),
                    widgets.VBox(
                        [
                            phoneTextBox,
                            facebookIdTextBox,
                            fbAccessTokenTextBox,
                            userType,
                        ],
                        layout=Layout(width="100%"),
                    ),
                ],
                layout=Layout(width="100%"),
            ),
            newMammutButton,
        ]
    )

    selectUserButton = widgets.Button(description="Seleccionar", button_style="primary")
    selectSupervisorButton = widgets.Button(
        description="Seleccionar", button_style="primary"
    )

    selectUserBox = widgets.HBox([personUserDropdown, selectUserButton])
    selectSupervisorBox = widgets.HBox([supervisorUserDropdown, selectSupervisorButton])

    user_tab = widgets.Tab(children=[selectUserBox, selectSupervisorBox, userBox])
    user_tab.set_title(0, "Seleccionar Persona Trabajando")
    user_tab.set_title(1, "Seleccionar supervisor")
    user_tab.set_title(2, "Crear Nuevo  Persona...")

    return (
        selectUserButton,
        selectSupervisorButton,
        newMammutButton,
        user_tab,
        form_values,
    )



from mammut.common.storage.storage_manager import StorageManager
import pandas as pd
from mammut.curriculum.course.course import Course
from mammut.assessment.assessment import Assessment
import json


class PackageProgress:
    """This class knows encapsulates the progress report functionality
    for the curriculum package document.

    The class contains a pandas.Dataframe which contains the curriculum
    progress in table form.

    The dataframe progressively appends new rows for each applied
    assessment. This allows to follow the progress over time.
    """

    _columns_names = [
        "course_id",
        "course_name",
        "course_attempt",
        "approved",
        "score",
        "assessment",
        "approved_problems",
        "failed_problems",
        "failed_problems_report",
    ]

    def __init__(
        self,
        curriculum: "Curriculum",
        storage_manager: StorageManager,
        progress_sheet_name: str,
        create_sheet: bool = False,
    ):
        self.progress_sheet_name = progress_sheet_name
        self._storage_manager = storage_manager
        self._curriculum_reference = curriculum
        self._dataframe: pd.DataFrame = None
        self._create_report_dataframe()
        if create_sheet:
            self._write_columns_titles()

    def _create_report_dataframe(self):
        self._dataframe = pd.DataFrame([], columns=self._columns_names)

    def write_progress(self, row):
        self._storage_manager.add_rows_to_spreadsheet(
            self._curriculum_reference.package_spreadsheet_id,
            self.progress_sheet_name,
            row.to_numpy().tolist(),
        )

    def _write_columns_titles(self):
        self._storage_manager.delete_sheet(
            self._curriculum_reference.package_spreadsheet_id,
            self.progress_sheet_name,
        )
        self._storage_manager.create_new_sheet(
            self._curriculum_reference.package_spreadsheet_id,
            self.progress_sheet_name,
        )
        self._storage_manager.update_row_from_spreadsheet(
            self._curriculum_reference.package_spreadsheet_id,
            self.progress_sheet_name,
            self._columns_names,
            "1",
            "A",
        )

    def add_progress_entry(self, course: Course, assessment: Assessment):
        row = pd.DataFrame(
            [
                [
                    str(course.course_id),
                    str(course.course_name),
                    str(course.last_attempt),
                    str(course.approved()),
                    str(course.last_score),
                    str(assessment.assessment_id),
                    "".join(str(assessment.approved_problems_id)),
                    "".join(str(assessment.failed_problems_id)),
                    json.dumps(
                        assessment.failed_problems_info,
                        sort_keys=False,
                        indent=4,
                    ),
                ]
            ],
            columns=self._columns_names,
        )
        self.write_progress(row)
        self._dataframe = self._dataframe.append(row)



from mammut.curriculum.models.hugging_face.pre_trained_model import (
    HuggingFacePreTrainedModel,
)
from transformers import BertModel, BertTokenizer, BertForTokenClassification
import mammut.curriculum.report as report
from mammut.common.corpus.corpus_map import CorpusMap
from typing import Dict
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class HuggingFaceBertBaseModel(HuggingFacePreTrainedModel):
    """This is a wrapper for the Bert Model transformer.

    https://huggingface.co/transformers/model_doc/bert.html#bertmodel

    """

    def __init__(
        self,
        parameters: Dict,
        course_id: int,
        lesson_id: int,
        mammut_session_context: MammutSessionContext,
        corpus_id: int,
        corpus_map: CorpusMap,
    ):
        """
       Args:
          parameters: The model parameters parsed JSON.
          course_id (int): the course ID to which this model belongs.
          lesson_id (int): the lesson ID to which this model belongs.
          mammut_session_context: Mammut session context with general information about
               current curriculum instance.
       """
        super(HuggingFaceBertBaseModel, self).__init__(
            parameters, course_id, lesson_id, mammut_session_context, corpus_id, corpus_map
        )

    def load_pretrained_models(self) -> None:
        """Load the pretrained model for this BERT wrapper. """
        self._model = BertModel.from_pretrained(self._name_or_path)
        self._tokenizer = BertTokenizer.from_pretrained(self._name_or_path)

    def train(self, mammut_session_context: MammutSessionContext, **kwargs):
        """Train the Bert base model in the lesson.

        Note: Nothing to train yet.
        Todo: Collaboration between Mammut Corpus (or other data source) and
            this model needs to be defined.
        """
        report.send_message(
            report.CurriculumGeneralDebugMessage(
                "Training HuggingFaceBertBase model"
            )
        )
        pass

    def get_corpus(
        self, corpus_id: int, corpus_map: CorpusMap,
    ):
        """
        Todo: Deprecated. Remove after new Hugging Face Model structure integration.
        """
        return corpus_map.get_corpus_by_id(corpus_id).get_events_as_strings()


class HuggingFaceBertTokenClassificationModel(HuggingFacePreTrainedModel):
    """This is a wrapper for the BertForTokenClassification model

    https://huggingface.co/transformers/model_doc/bert.html#bertfortokenclassification

    """

    def __init__(
        self,
        parameters: Dict,
        course_id: int,
        lesson_id: int,
        mammut_session_context: MammutSessionContext,
        corpus_id: int,
        corpus_map: CorpusMap,
    ):
        """
        Args:
          parameters: The model parameters parsed JSON.
          course_id (int): the course ID to which this model belongs.
          lesson_id (int): the lesson ID to which this model belongs.
          mammut_session_context: Mammut session context with general information about
               current curriculum instance.
       """
        super(HuggingFaceBertTokenClassificationModel, self).__init__(
            parameters, course_id, lesson_id, mammut_session_context, corpus_id, corpus_map
        )

    def load_pretrained_models(self) -> None:
        """Load the pretrained model for this BERT wrapper. """
        self._model = BertForTokenClassification.from_pretrained(
            self._name_or_path
        )
        self._tokenizer = BertTokenizer.from_pretrained(self._name_or_path)

    def train(self, mammut_session_context: MammutSessionContext, **kwargs):
        """Train this NER model in the lesson.

        Note: nothing to train yet.
        Todo: Collaboration between Mammut Corpus (or other data source) and
            this model needs to be defined.
        """
        report.send_message(
            report.CurriculumGeneralDebugMessage(
                "Training HuggingFaceBertTokenClassificationModel model"
            )
        )
        pass

    def get_corpus(
        self, corpus_id: int, corpus_map: CorpusMap,
    ):
        """
        Todo: Deprecated. Remove after new Hugging Face Model structure integration.
        """
        return corpus_map.get_corpus_by_id(corpus_id).get_events_as_strings()



import enum
from mammut.assessment import (
    sentence_piece_problem,
    mock_problem,
    tokenizer_problems,
    token_classification_problems,
)
from typing import List, Tuple


class ProblemType(enum.Enum):
    """Enumeration for checklist test types.

    Enumeration values are custom classes for problem types.

    Enumeration naming follows a convention:

        Problem category name + "_" +  Model name

    For example: For a minimal functionality test problem
    implemented for Sentencepiece, the enum name will be:

        MFT_Sentencepiece
    """

    """
        Minimal functionality test problems
    """
    MFT_SentencePiece = sentence_piece_problem.SentencepieceProblemMFT
    MFT_SpecialToken = tokenizer_problems.SpecialTokenProblemMFT
    MFT_TokenClassification = token_classification_problems.RobertaTokenClassificationProblemMFT

    """
        Invariance functionality test
    """
    INV_SentencePiece = sentence_piece_problem.SentencepieceProblemINV

    """
        General category. (Problem not tied to any specific category)
    """
    General_Mock = mock_problem.MockProblem

    @classmethod
    def list(cls) -> List[Tuple]:
        """Returns a list of pairs for the enumeration names and values.
        """
        return list(map(lambda e: (e.name, e.value), cls))

    @classmethod
    def names(cls) -> List[str]:
        """Returns the list of enumeration names
        """
        return list(map(lambda e: e.name, cls))



from collections import UserDict, UserList
from mammut.common.storage.storage_manager import StorageManager
from typing import List, Dict
import pandas as pd


class FunctionalParameterValue:
    def __init__(self, value: str, parameter):
        self.value = value
        self.parameter = parameter


class PrototypicalFunctionalParameterValue(FunctionalParameterValue):
    def __init__(
        self,
        value: str,
        value_type: str,
        show: bool,
        definition: str,
        example: str,
        parameter,
    ):
        FunctionalParameterValue.__init__(self, value, parameter)
        self.value_type = value_type
        self.show = show
        self.definition = definition
        self.example = example


class FunctionalParameterDict(UserDict):
    def __init__(self, standard, spreadsheet_id):
        UserDict.__init__(self)
        self.standard = standard
        self.spreadsheet_id = spreadsheet_id
        self.valid = True
        self.errors = []


class FunctionalParameter:
    def __init__(self, spreadsheet_id: str, standard):
        self.spreadsheet_id = spreadsheet_id
        self.standard = standard
        self.parameter_values = {}


class PrototypicalFunctionalParameter(FunctionalParameter):
    SHEET_NAME = "parameters"
    # PARAMETERS COLUMN NAMES
    CLASS_COLUMN_NAME = "class"
    VALUE_COLUMN_NAME = "value"
    VALUE_TYPE_COLUMN_NAME = "value-type"
    SHOW_COLUMN_NAME = "show"
    DEFINITION_COLUMN_NAME = "definition"
    EXAMPLE_COLUMN_NAME = "example"
    storage_manager = StorageManager()
    DEFAULT_START_CELL = "A1"
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "1"
    DEFAULT_END_CELL = "F503"
    DEFAULT_END_COLUM = "F"

    LOADED_PARAMETERS = {}

    def __init__(self, p_class, group, spreadsheet_id: str, standard):
        FunctionalParameter.__init__(self, spreadsheet_id, standard)
        self.p_class = p_class
        self.group = group
        for ig in self.group.iterrows():
            pv = PrototypicalFunctionalParameterValue(
                ig[1][PrototypicalFunctionalParameter.VALUE_COLUMN_NAME],
                ig[1][PrototypicalFunctionalParameter.VALUE_TYPE_COLUMN_NAME],
                ig[1][PrototypicalFunctionalParameter.SHOW_COLUMN_NAME] != "",
                ig[1][PrototypicalFunctionalParameter.DEFINITION_COLUMN_NAME],
                ig[1][PrototypicalFunctionalParameter.EXAMPLE_COLUMN_NAME],
                self,
            )
            self.parameter_values[pv.value] = pv

    @staticmethod
    def load(spreadsheet_id: str, standard):
        res = None
        if spreadsheet_id in PrototypicalFunctionalParameter.LOADED_PARAMETERS:
            res = PrototypicalFunctionalParameter.LOADED_PARAMETERS[
                spreadsheet_id
            ]
        else:
            res = FunctionalParameterDict(standard, spreadsheet_id)
            res.range_name = (
                PrototypicalFunctionalParameter.SHEET_NAME
                + "!"
                + PrototypicalFunctionalParameter.DEFAULT_START_CELL
            )

            res.data_frame = PrototypicalFunctionalParameter.storage_manager.get_spreadsheet_as_dataframe(
                res.spreadsheet_id,
                PrototypicalFunctionalParameter.SHEET_NAME,
                PrototypicalFunctionalParameter.DEFAULT_START_COLUMN,
                PrototypicalFunctionalParameter.DEFAULT_START_ROW,
                PrototypicalFunctionalParameter.DEFAULT_END_COLUM,
            )

            if res.data_frame is not None:
                res.group_by_object = res.data_frame.groupby(
                    [PrototypicalFunctionalParameter.CLASS_COLUMN_NAME]
                )
                for g in res.group_by_object:
                    prot = PrototypicalFunctionalParameter(
                        g[0], g[1], spreadsheet_id, standard
                    )
                    res[prot.p_class] = prot
            else:
                res.valid = False
                res.errors.append(f"The parameters sheet is not present.")
            PrototypicalFunctionalParameter.LOADED_PARAMETERS[
                spreadsheet_id
            ] = res
        return res


class FunctionalLemmaDefinition:
    def __init__(self, lemma, index):
        self.lemma = lemma
        self.index = index

    def get_description(self):
        raise NotImplementedError()


class PrototypicalFunctionalLemmaDefinition(FunctionalLemmaDefinition):
    # MODEL COLUMN NAMES
    INDEX_COLUMN_NAME = "index"
    HIDE_COLUMN_NAME = "hide"
    PROTOTYPICAL_EXAMPLE_COLUMN_NAME = "prototypical-example"
    COMMENTS_COLUMN_NAME = "comments"
    ETYMOLOGICAL_INDEX_COLUMN_NAME = "etymological-index"
    ETYMOLOGICAL_INFO_COLUMN_NAME = "etymological-info"

    storage_manager = StorageManager()
    DEFAULT_START_CELL = "A1"
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "1"
    DEFAULT_END_CELL = "S503"
    DEFAULT_END_COLUMN = "S"

    LOADED_LEMMAS = {}

    COMMENTS_DEFAULT = "AUN NO ESTA LISTO!!!"
    ETYMOLOGICAL_INDEX_DEFAULT = "-"
    ETYMOLOGICAL_INFO_DEFAULT = "-"

    def __init__(self, lemma, row, spreadsheet_id, parameters, standard):
        FunctionalLemmaDefinition.__init__(
            self,
            lemma,
            row[PrototypicalFunctionalLemmaDefinition.INDEX_COLUMN_NAME],
        )
        self.row = row
        self.prototypical_example = self.row[
            PrototypicalFunctionalLemmaDefinition.PROTOTYPICAL_EXAMPLE_COLUMN_NAME
        ]
        self.comments = PrototypicalFunctionalLemmaDefinition.COMMENTS_DEFAULT
        if (
            PrototypicalFunctionalLemmaDefinition.COMMENTS_COLUMN_NAME
            in self.row
        ):
            self.comments = self.row[
                PrototypicalFunctionalLemmaDefinition.COMMENTS_COLUMN_NAME
            ]
        self.etymological_index = (
            PrototypicalFunctionalLemmaDefinition.ETYMOLOGICAL_INDEX_DEFAULT
        )
        if (
            PrototypicalFunctionalLemmaDefinition.ETYMOLOGICAL_INDEX_COLUMN_NAME
            in self.row
        ):
            self.etymological_index = self.row[
                PrototypicalFunctionalLemmaDefinition.ETYMOLOGICAL_INDEX_COLUMN_NAME
            ]
        self.etymological_info = (
            PrototypicalFunctionalLemmaDefinition.ETYMOLOGICAL_INFO_DEFAULT
        )
        if (
            PrototypicalFunctionalLemmaDefinition.ETYMOLOGICAL_INFO_COLUMN_NAME
            in self.row
        ):
            self.etymological_info = self.row[
                PrototypicalFunctionalLemmaDefinition.ETYMOLOGICAL_INFO_COLUMN_NAME
            ]
        self.spreadsheet_id = spreadsheet_id
        self.parameters = parameters
        self.standard = standard
        self.valid = True
        self.errors = []
        for p in self.parameters:
            paramd = self.parameters[p]
            if p in self.row:
                pv = self.row[p]
                if pv not in paramd.parameter_values:
                    self.valid = False
                    self.errors.append(
                        f"Lemma: {lemma} in spreadsheet: {spreadsheet_id} for index: {self.index} contain a invalid value for parameter: {p}. Value: {pv}"
                    )
                setattr(self, p, pv)
            else:
                self.valid = False
                self.errors.append(
                    f"Lemma: {lemma} in spreadsheet: {spreadsheet_id} for index: {self.index} does not contain parameter: {p}"
                )

    def get_description(self):
        res = f"Ejemplo prototipico: {self.prototypical_example}\nComentarios: {self.comments}"
        for p in self.parameters:
            pv = getattr(self, p)
            paramd = self.parameters[p]
            descriptor = paramd.parameter_values[pv]
            if descriptor.show:
                res += f"\n{descriptor.value_type}: {pv}"
        return res

    @staticmethod
    def get_definitions(
        lemma,
        model_parameters_spreadsheet_id: str,
        model_spreadsheet_id: str,
        standard,
    ):
        parameters = PrototypicalFunctionalParameter.load(
            model_parameters_spreadsheet_id, standard
        )
        res = None
        loading_id = f"{model_spreadsheet_id}---{lemma}"
        if loading_id in PrototypicalFunctionalLemmaDefinition.LOADED_LEMMAS:
            res = PrototypicalFunctionalLemmaDefinition.LOADED_LEMMAS[
                loading_id
            ]
        else:
            res = FunctionalLemmaDefinitionList(standard, model_spreadsheet_id)
            if not parameters.valid:
                res.valid = False
                res.errors = res.errors + parameters.errors
            res.range_name = (
                lemma
                + "!"
                + PrototypicalFunctionalLemmaDefinition.DEFAULT_START_CELL
            )

            res.data_frame = PrototypicalFunctionalLemmaDefinition.storage_manager.get_spreadsheet_as_dataframe(
                res.spreadsheet_id,
                lemma,
                PrototypicalFunctionalLemmaDefinition.DEFAULT_START_COLUMN,
                PrototypicalFunctionalLemmaDefinition.DEFAULT_START_ROW,
                PrototypicalFunctionalLemmaDefinition.DEFAULT_END_COLUMN,
            )

            if (
                PrototypicalFunctionalLemmaDefinition.HIDE_COLUMN_NAME
                in res.data_frame
            ):
                res.data_frame = res.data_frame[
                    res.data_frame[
                        PrototypicalFunctionalLemmaDefinition.HIDE_COLUMN_NAME
                    ]
                    == ""
                ]
            if res.data_frame is not None:
                for r in res.data_frame.iterrows():
                    defi = PrototypicalFunctionalLemmaDefinition(
                        lemma, r[1], model_spreadsheet_id, parameters, standard
                    )
                    res.append(defi)
                    if not defi.valid:
                        res.valid = False
                        res.errors = res.errors + defi.errors

            else:
                res.valid = False
                res.errors.append(
                    f"The protypical definitions sheet is not present."
                )
            PrototypicalFunctionalLemmaDefinition.LOADED_LEMMAS[
                loading_id
            ] = res
        return res


class FunctionalLemmaDescriptor:
    LEMMA_COLUMN_NAME = "lemma"
    HIDE_COLUMN_NAME = "hide"
    WORD_TYPE_COLUMN_NAME = "word-type"
    REGIONAL_SETTINGS_LEMMA_COLUMN_NAME = "regional-settings-lemma"
    POS_PARADIGM_COLUMN_NAME = "pos-paradigm"
    MODEL_TYPE_COLUMN_NAME = "model-type"
    MODEL_PARAMETERS_SPREADSHEET_ID_COLUMN_NAME = (
        "model-parameters-spreadsheet-id"
    )
    MODEL_SPREADSHEET_ID_COLUMN_NAME = "model-spreadsheet-id"
    MORPH_PARADIGM_COLUMN_NAME = "morph-paradigm"
    BOUND_MORPH_COLUMN_NAME = "bound-morph"
    MISS_PARADIGM_COLUMN_NAME = "miss-paradigm"
    FET_PARADIGM_COLUMN_NAME = "fet-paradigm"
    FEATURE_CLASS_COLUMN_NAME = "feature-class"
    FEATURE_ID_COLUMN_NAME = "feature-id"
    FEATURE_POS_COLUMN_NAME = "feature-pos-tag"
    DEFAULT_START_CELL = "A1"
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "1"
    DEFAULT_END_CELL = "O"

    def __init__(self, dataframe_row, standard):
        self.lemma = (
            dataframe_row[FunctionalLemmaDescriptor.LEMMA_COLUMN_NAME]
            .lower()
            .strip()
        )
        self.word_type = dataframe_row[
            FunctionalLemmaDescriptor.WORD_TYPE_COLUMN_NAME
        ]
        self.regional_settings_lemma = dataframe_row[
            FunctionalLemmaDescriptor.REGIONAL_SETTINGS_LEMMA_COLUMN_NAME
        ]
        self.pos_paradigm = dataframe_row[
            FunctionalLemmaDescriptor.POS_PARADIGM_COLUMN_NAME
        ]
        self.model_type = dataframe_row[
            FunctionalLemmaDescriptor.MODEL_TYPE_COLUMN_NAME
        ]
        self.model_parameters_spreadsheet_id = dataframe_row[
            FunctionalLemmaDescriptor.MODEL_PARAMETERS_SPREADSHEET_ID_COLUMN_NAME
        ]
        self.model_spreadsheet_id = dataframe_row[
            FunctionalLemmaDescriptor.MODEL_SPREADSHEET_ID_COLUMN_NAME
        ]
        self.morph_paradigm = dataframe_row[
            FunctionalLemmaDescriptor.MORPH_PARADIGM_COLUMN_NAME
        ]
        self.is_bound = bool(
            dataframe_row[FunctionalLemmaDescriptor.BOUND_MORPH_COLUMN_NAME]
        )
        self.miss_paradigm = dataframe_row[
            FunctionalLemmaDescriptor.MISS_PARADIGM_COLUMN_NAME
        ]
        self.fet_paradigm = dataframe_row[
            FunctionalLemmaDescriptor.FET_PARADIGM_COLUMN_NAME
        ]
        self.feature_class = dataframe_row[
            FunctionalLemmaDescriptor.FEATURE_CLASS_COLUMN_NAME
        ]
        self.feature_id = dataframe_row[
            FunctionalLemmaDescriptor.FEATURE_ID_COLUMN_NAME
        ]
        self.feature_pos = dataframe_row[
            FunctionalLemmaDescriptor.FEATURE_POS_COLUMN_NAME
        ]
        self.standard = standard
        self.id = self.build_id(
            self.lemma, self.regional_settings_lemma, self.pos_paradigm
        )
        self.implicit_condition_id = None
        if not self.is_explicit():
            self.implicit_condition_id = self.build_implicit_condition_id(
                self.feature_class,
                self.feature_id,
                self.feature_pos,
                self.regional_settings_lemma,
            )
        self.definitions = []
        self.undefined = False
        self.valid = True
        self.errors = []
        if self.model_type == "prototypical":
            self.definitions = PrototypicalFunctionalLemmaDefinition.get_definitions(
                self.lemma,
                self.model_parameters_spreadsheet_id,
                self.model_spreadsheet_id,
                self.standard,
            )
            if not self.definitions.valid:
                self.valid = False
                self.errors = self.errors + self.definitions.errors
        elif (
            self.model_type == ""
            and self.model_parameters_spreadsheet_id == ""
            and self.model_spreadsheet_id == ""
        ):
            self.undefined = True
        else:
            self.valid = False
            self.errors.append(f"The model type is invalid.")

    def is_explicit(self):
        return not self.feature_class

    @staticmethod
    def build_implicit_condition_id(
        feature_class: str,
        feature_id: str,
        feature_pos: str,
        regional_settings_lemma: str,
    ):
        return f"{feature_class}-{feature_id}-{feature_pos}-{regional_settings_lemma}"

    @staticmethod
    def build_id(lemma: str, regional_settings_lemma: str, pos_paradigm: str):
        return f"{lemma}-{regional_settings_lemma}-{pos_paradigm}"


class FunctionalLemmaDefinitionList(UserList):
    def __init__(self, standard, spreadsheet_id, initlist=None):
        UserList.__init__(self, initlist)
        self.standard = standard
        self.spreadsheet_id = spreadsheet_id
        self.valid = True
        self.errors = []


class FunctionalModelComparison:
    """Compare a prototypical modeling with another modeling of the same type."""

    storage_manager = StorageManager()

    def __init__(self, item_name: str, sheet_id: str):
        self.item_name = item_name
        self.sheet_id = sheet_id
        self.start_position = ("A", "1")
        self.end_position = "I"
        self.pivot = False
        self.model = self._get_model_from_sheets(self.sheet_id, self.item_name)

    def _get_model_from_sheets(
        self, sheet_id: str, item_name: str
    ) -> pd.DataFrame:
        """Get model from sheet_ID."""
        return FunctionalModelComparison.storage_manager.get_spreadsheet_as_dataframe(
            sheet_id,
            item_name,
            self.start_position[0],
            self.start_position[1],
            self.end_position,
            self.pivot,
        )

    def _merge_df(
        self, model1: pd.DataFrame, model2: pd.DataFrame, features: List
    ) -> pd.DataFrame:
        """Merge two dataframes from a join operation based on the features parameter."""
        return model1.merge(
            model2,
            how="inner",
            indicator=True,
            on=features,
            suffixes=["_model", "_alt"],
        )

    def duplicates_in_model(self, features: List) -> pd.DataFrame:
        """Find repeated modeling in the prototypical model."""
        try:
            return self.model[self.model.loc[:, features].duplicated(keep=False)]
        except KeyError as e:
            return f"Error. Column not found. See {str(e)}"

    def non_duplicates_in_model(self, features: List) -> pd.DataFrame:
        """Find an example of each non-duplicate modeling in the model."""
        try:
            return self.model[~self.model.loc[:, features].duplicated()]
        except KeyError as e:
            return f"Error. Column not found. See {str(e)}"

    def compare_to(self, item_name: str, features: List) -> pd.DataFrame:
        """It compares the current model with another model and returns the modelings in common."""
        self.alternative_model = self._get_model_from_sheets(
            self.sheet_id, item_name
        )
        model1 = self.model[~self.model.loc[:, features].duplicated()]
        model2 = self.alternative_model[
            ~self.alternative_model.loc[:, features].duplicated()
        ]
        return self._merge_df(model1, model2, features)

    def compare_to_list(
        self, models: List, features: List, join: bool = False
    ) -> Dict[str, pd.DataFrame]:
        """It compares the current model with another models (from a list) and returns the modelings in common."""
        model_dfs = {}
        original_dfs = []
        for item_name in models:
            alt_model = self._get_model_from_sheets(self.sheet_id, item_name)
            model1 = self.model[
                ~self.model.loc[:, features].duplicated()
            ].copy()
            model1["item_name"] = self.item_name
            model2 = alt_model[~alt_model.loc[:, features].duplicated()].copy()
            model2["item_name"] = item_name
            merged_model = self._merge_df(model1, model2, features)
            model_dfs[item_name] = merged_model
            original_dfs.append(model2)

        if join:
            joined_dfs = pd.concat(original_dfs)
            joined = joined_dfs[
                ~joined_dfs.loc[:, features].duplicated().copy()
            ]
            merged_joined = self._merge_df(self.model, joined, features)
            model_dfs["__all__"] = merged_joined
        return model_dfs



# coding=utf-8
import os
import configparser
from distutils.util import strtobool
from pkg_resources import resource_string, resource_filename

__all__ = [
    "configurable",
    "imaging",
    "lexical_discrete_choice",
    "lexical_relevance",
    "node",
    "parameter_node",
    "reparameterization",
    "trainable_node",
    "trainer",
    "transition_system",
    "worker_node",
]

MODELS_CONFIGURATIONS = configparser.ConfigParser()
MODELS_CONFIGURATIONS.read_string(
    resource_string("mammut.resources", "models.ini").decode("utf-8")
)
MODELS_CONFIGURATIONS.read(["data/models.ini", "models.ini"])
MODELS_CONFIGURATIONS_FILE_NAME = resource_filename("mammut.resources", "models.ini")
if os.path.exists("data/models.ini"):
    MODELS_CONFIGURATIONS_FILE_NAME = "data/models.ini"
if os.path.exists("models.ini"):
    MODELS_CONFIGURATIONS_FILE_NAME = "models.ini"

NGRAM_VIEWER_SMOOTHING = MODELS_CONFIGURATIONS["NGRAM_VIEWER"]["smoothing"]
NGRAM_VIEWER_YEAR_START = int(MODELS_CONFIGURATIONS["NGRAM_VIEWER"]["year_start"])
NGRAM_VIEWER_YEAR_END = int(MODELS_CONFIGURATIONS["NGRAM_VIEWER"]["year_end"])
NGRAM_VIEWER_CASE_INSENSITIVE = strtobool(
    MODELS_CONFIGURATIONS["NGRAM_VIEWER"]["case_insensitive"]
)
LEXICAL_RELEVANCE_INDEX_ESTIMATOR_USAGE_SOURCE = MODELS_CONFIGURATIONS[
    "LEXICAL_RELEVANCE_INDEX_ESTIMATOR"
]["usage_source"]
LEXICAL_RELEVANCE_INDEX_ESTIMATOR_CONSOLIDATE_STATS = strtobool(
    MODELS_CONFIGURATIONS["LEXICAL_RELEVANCE_INDEX_ESTIMATOR"]["consolidate_stats"]
)
LEXICAL_RELEVANCE_INDEX_ESTIMATOR_AUMENTATION = int(
    MODELS_CONFIGURATIONS["LEXICAL_RELEVANCE_INDEX_ESTIMATOR"]["aumentation"]
)

DISTRIBUTED_TRAINING = strtobool(MODELS_CONFIGURATIONS["TRAINING"]["distributed"])



# coding=utf-8


class SheetsStorageBase(object):
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "1"
    DEFAULT_END_RANGE = "Z"

    def __init__(self, storage_path=""):
        self.storage_path = storage_path

    def get_spreadsheet_as_dataframe(
        self,
        spreadsheet_id: str,
        sheet_name,
        start_column: str,
        start_row: str,
        end_range: str,
        transpose: bool = False,
    ):
        pass

    def add_rows_to_spreadsheet(
        self,
        spreadsheet_id: str,
        sheet_name: str,
        rows,
        start_row=DEFAULT_START_ROW,
        start_col=DEFAULT_START_COLUMN,
        end_range=DEFAULT_END_RANGE,
    ):
        pass

    def append_row_to_sheet(
        self,
        spreadsheet_id: str,
        sheet_name: str,
        row,
        start_row=DEFAULT_START_ROW,
        start_col=DEFAULT_START_COLUMN,
        end_range=DEFAULT_END_RANGE,
    ):
        pass

    def create_new_sheet(self, spreadsheet_id: str, sheet_name: str):
        pass

    def create_temporal_new_sheet(self, spreadsheet_id: str, base_sheet_title: str):
        pass

    def get_temporal_sheets_from_new_to_old(self, spreadsheet_id, base_sheet_title):
        pass

    def get_sheet_id(self, spreadsheet_id, identifier):
        pass

    def update_row_from_spreadsheet(
        self, spreadsheet_id, sheet_name, row, start_row, start_col
    ):
        pass

    def update_rows_from_spreadsheet(
        self, spreadsheet_id, sheet_name, rows, start_row, start_col
    ):
        pass

    def delete_rows_from_spreadsheet(
        self, spreadsheet_id, sheet_name, start_index, end_index
    ):
        pass

    def delete_sheet(self, spreadsheet_id, sheet_name):
        pass


class SlidesStorageBase(object):
    slides_metadata_re_string = "(sheet-id|Sheet-id|SHEET-ID)[ ]*->[ ]*([a-zA-Z0-9-_]+)"

    def __init__(self, storage_path=""):
        self.storage_path = storage_path

    def get_slide_elements(self, presentation_id):
        pass

    def get_slides_metadata_re_string(self):
        return self.slides_metadata_re_string



# coding=utf-8

import redis
import functools
from abc import *
from mammut.models import MODELS_CONFIGURATIONS, MODELS_CONFIGURATIONS_FILE_NAME

CONFIGURATION_PROPERTIES_DEFAULT_VALUES = {}
CONFIGURATION_PROPERTIES_PARSE_FUNCTION = {}


def _get_configuration_property(self, class_name, property_name):
    configuration_path = self._get_configuration_path()
    default_value = CONFIGURATION_PROPERTIES_DEFAULT_VALUES[class_name][property_name]
    parser_function = CONFIGURATION_PROPERTIES_PARSE_FUNCTION[class_name][property_name]
    if self._redisProxy != None:
        if not self._redisProxy.hexists(configuration_path, property_name):
            _set_configuration_property(self, default_value, property_name)
        if parser_function != None:
            return parser_function(
                self._redisProxy.hget(configuration_path, property_name)
            )
        else:
            return self._redisProxy.hget(configuration_path, property_name)
    else:
        if not MODELS_CONFIGURATIONS.has_option(configuration_path, property_name):
            _set_configuration_property(self, default_value, property_name)
        if parser_function != None:
            return parser_function(
                MODELS_CONFIGURATIONS[configuration_path][property_name]
            )
        else:
            return MODELS_CONFIGURATIONS[configuration_path][property_name]


def _set_configuration_property(self, value, property_name):
    configuration_path = self._get_configuration_path()
    if self._redisProxy != None:
        self._redisProxy.hset(configuration_path, property_name, value)
    else:
        if not MODELS_CONFIGURATIONS.has_section(configuration_path):
            MODELS_CONFIGURATIONS[configuration_path] = {}
        MODELS_CONFIGURATIONS[configuration_path][property_name] = str(value)
        with open(MODELS_CONFIGURATIONS_FILE_NAME, "w") as configfile:
            MODELS_CONFIGURATIONS.write(configfile)


class Configurable(metaclass=ABCMeta):
    def __init__(
        self, configuration_path, configuration_name, redis_connection_pool=None
    ):
        """"""
        self._configuration_parent = configuration_path
        self._configuration = configuration_name
        self._redisProxy = None
        if redis_connection_pool != None:
            self._redisProxy = redis.Redis(connection_pool=redis_connection_pool)

    def _get_configuration_path(self):
        return self._configuration_parent + "." + self._configuration

    @staticmethod
    def _add_configuration_property(
        class_name,
        property_name,
        default_value,
        parser_function=None,
        property_doc=None,
    ):
        if class_name not in CONFIGURATION_PROPERTIES_DEFAULT_VALUES:
            CONFIGURATION_PROPERTIES_DEFAULT_VALUES[class_name] = {}
            CONFIGURATION_PROPERTIES_PARSE_FUNCTION[class_name] = {}
        CONFIGURATION_PROPERTIES_DEFAULT_VALUES[class_name][
            property_name
        ] = default_value
        CONFIGURATION_PROPERTIES_PARSE_FUNCTION[class_name][
            property_name
        ] = parser_function
        return property(
            fget=functools.partial(
                _get_configuration_property,
                class_name=class_name,
                property_name=property_name,
            ),
            fset=functools.partial(
                _set_configuration_property, property_name=property_name
            ),
            doc=property_doc,
        )



from abc import ABC, abstractmethod
from typing import List


class CorpusOperations(ABC):
    """Interface shared across all corpus types
    to offer operations:

     - Knowing whether the corpus can be indexed or not.
     - Knowing whether the corpus can be annotated or not.
     - Get the corpus events as a List of strings: this can be useful for
        training third party models that receives text as a list of string.
    """
    @property
    @abstractmethod
    def indexable(self) -> bool:
        pass

    @property
    @abstractmethod
    def annotable(self) -> bool:
        pass

    @property
    @abstractmethod
    def get_events_as_strings(self) -> List[str]:
        """
        Todo: As the Curriculum development takes shape, this method should be
            deprecated and removed. Models will eventually require a structured
            data input that will be implemented in corpus that are compliant with
            the incorporated models.
        :return:
        """
        pass



import itertools

import pandas as pd

from mammut.common.basic_tokenizer import BasicTokenizer
from mammut.common.corpus.synthetic_corpus import SyntheticCorpus
from mammut.common.corpus.images_corpus import ImagesCorpus
from mammut.common.corpus.crawler_corpus import CrawlerCorpus
from mammut.common.corpus.dictionary_corpus import DictionaryCorpus
from mammut.common.corpus.natural_corpus import NaturalCorpus
from mammut.common.corpus.issue_corpus import IssueCorpus
from mammut.common.corpus.mammut_corpus import MammutCorpus
from mammut.common.lexicon.dictionary import Dictionary
from mammut.common.lexicon.linguistic_standard import LinguisticStandard
from mammut.common.storage.storage_manager import StorageManager
from mammut.common.catalog.catalog_manager import Catalog
from typing import List
from mammut.common.corpus.hugging_face_corpus import *
import json as json_module


class CorpusMap:
    """The Corpus Map is a map of all the corpus
    that can be used to train a bot.

    # TODO: Document class methods.
    """

    DEFAULT_SHEET_TITTLE = "corpus_map"
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "2"
    DEFAULT_END_RANGE = "N"
    DEFAULT_ID_KEY = "id"
    DEFAULT_MAIN_SHEET_KEY = "main_sheet"
    DEFAULT_EXTENSION_SHEET_KEY = "extension_sheet"
    DEFAULT_VARIABLES_SHEET_KEY = "variables_sheet"
    DEFAULT_TYPE_KEY = "corpus_type"
    DEFAULT_REGIONAL_SETTINGS_KEY = "regional_settings"
    DEFAULT_ANNOTABLE_KEY = "annotable"
    DEFAULT_ONTOLOGY_INSTANCES_KEY = "ontology_instances"
    DEFAULT_ONTOLOGY_DEFAULTS_KEY = "ontology_defaults"
    DEFAULT_ONTOLOGY_VERTICES_KEY = "ontology_vertices"
    DEFAULT_ONTOLOGY_EDGES_KEY = "ontology_edges"
    DEFAULT_ONTOLOGY_PROPERTIES_KEY = "ontology_properties"
    DEFAULT_CONFIGURATION_JSON_KEY = "configuration_json"
    DEFAULT_HIDDEN_KEY = "hidden"
    DEFAULT_MAMMUT_TYPE = "M"
    DEFAULT_CRAWLER_TYPE = "C"
    DEFAULT_NATURAL_TYPE = "N"
    DEFAULT_IMAGES_TYPE = "IM"
    DEFAULT_ISSUES_TYPE = "I"
    DEFAULT_DICTIONARY_TYPE = "D"
    DEFAULT_SYNTHETIC_TYPE = "S"
    DEFAULT_HUGGINGFACE_FIXED_TYPE = "HF_FIXED"
    DEFAULT_NONE_KEY = "None"
    storage_manager = StorageManager()

    def __init__(
        self,
        spreadsheet_id: str,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        standard: LinguisticStandard,
        catalog: Catalog,
        update_status_text_box=None,
    ):

        self.map = self.storage_manager.get_spreadsheet_as_dataframe(
            spreadsheet_id,
            CorpusMap.DEFAULT_SHEET_TITTLE,
            CorpusMap.DEFAULT_START_COLUMN,
            CorpusMap.DEFAULT_START_ROW,
            CorpusMap.DEFAULT_END_RANGE,
        )
        if CorpusMap.DEFAULT_HIDDEN_KEY in self.map:
            self.map = self.map.loc[
                (self.map[CorpusMap.DEFAULT_HIDDEN_KEY] == "")
            ]
        self.mammut_corpus_list = []
        self.crawler_corpus_list = []
        self.natural_corpus_list = []
        self.images_corpus_list = []
        self.dictionary_corpus_list = []
        self.issues_corpus_list = []
        self.synthetic_corpus_list = []
        self.hugging_face_corpus_list: List[HuggingFaceDatasetCorpus] = []
        self._corpus_table_by_id: Dict = {}
        self.spreadsheet_id = spreadsheet_id
        self.tokenizer = tokenizer
        self.dictionary = dictionary
        self.standard = standard
        self.catalog = catalog
        self.update_status_text_box = update_status_text_box
        self.switch_load = {
            CorpusMap.DEFAULT_MAMMUT_TYPE: self.create_mammut_corpus,
            CorpusMap.DEFAULT_CRAWLER_TYPE: self.create_crawler_corpus,
            CorpusMap.DEFAULT_NATURAL_TYPE: self.create_natural_corpus,
            CorpusMap.DEFAULT_IMAGES_TYPE: self.create_images_corpus,
            CorpusMap.DEFAULT_ISSUES_TYPE: self.create_issue_corpus,
            CorpusMap.DEFAULT_SYNTHETIC_TYPE: self.create_synthetic_corpus,
            CorpusMap.DEFAULT_HUGGINGFACE_FIXED_TYPE: self.create_hugging_face_fixed_corpus,
        }
        self.switch_get = {
            CorpusMap.DEFAULT_MAMMUT_TYPE: self.mammut_corpus_list,
            CorpusMap.DEFAULT_CRAWLER_TYPE: self.crawler_corpus_list,
            CorpusMap.DEFAULT_NATURAL_TYPE: self.natural_corpus_list,
            CorpusMap.DEFAULT_IMAGES_TYPE: self.images_corpus_list,
            CorpusMap.DEFAULT_DICTIONARY_TYPE: self.dictionary_corpus_list,
            CorpusMap.DEFAULT_ISSUES_TYPE: self.issues_corpus_list,
            CorpusMap.DEFAULT_SYNTHETIC_TYPE: self.synthetic_corpus_list,
            CorpusMap.DEFAULT_HUGGINGFACE_FIXED_TYPE: self.hugging_face_corpus_list,
        }
        self.load()
        self.load_dictionary_corpus()

    def _update_corpus_table_by_id(self, corpus_id: int, corpus_reference):
        """Adds the corpus to the corpus table with corpus ID keys.

        The ID of the corpus is being received because there's no a common
        generalization of corpus data attributes.

        Args:
            corpus_id: The corpus ID for the corpus_reference.
            corpus_reference: the corpus object reference to be stored as value
                in the table.
        """
        self._corpus_table_by_id[int(corpus_id)] = corpus_reference

    def load(self):
        i = 0
        for type in self.map[CorpusMap.DEFAULT_TYPE_KEY]:
            annotable = not (
                str.strip(self.map[CorpusMap.DEFAULT_ANNOTABLE_KEY][i]) == ""
            )
            self.switch_load[type](
                self.map[CorpusMap.DEFAULT_ID_KEY][i],
                self.map[CorpusMap.DEFAULT_MAIN_SHEET_KEY][i],
                self.map[CorpusMap.DEFAULT_EXTENSION_SHEET_KEY][i],
                self.map[CorpusMap.DEFAULT_VARIABLES_SHEET_KEY][i],
                self.map[CorpusMap.DEFAULT_REGIONAL_SETTINGS_KEY][i],
                self.map[CorpusMap.DEFAULT_CONFIGURATION_JSON_KEY][i],
                annotable,
                self.map[CorpusMap.DEFAULT_ONTOLOGY_INSTANCES_KEY][i],
                self.map[CorpusMap.DEFAULT_ONTOLOGY_DEFAULTS_KEY][i],
                self.map[CorpusMap.DEFAULT_ONTOLOGY_VERTICES_KEY][i],
                self.map[CorpusMap.DEFAULT_ONTOLOGY_EDGES_KEY][i],
                self.map[CorpusMap.DEFAULT_ONTOLOGY_PROPERTIES_KEY][i],
            )
            i += 1

    def load_dictionary_corpus(self):
        lemmas_entries_by_regional_settings_lemma = itertools.groupby(
            self.dictionary.lemmas_entries, lambda e: e.regional_settings_lemma
        )
        for (
            regional_settings_lemma,
            lemmas_entries,
        ) in lemmas_entries_by_regional_settings_lemma:
            (
                sheet_rows,
                hash_lemma_dict,
            ) = DictionaryCorpus.transform_dictinary_to_sheet_rows(
                self.dictionary
            )
            corpus_dataframe = pd.DataFrame(
                sheet_rows[1:], columns=sheet_rows[0]
            )
            corpus_dataframe["sub_id"] = corpus_dataframe["sub_id"].apply(str)
            corpus_dataframe["id"] = corpus_dataframe["id"].apply(str)
            corpus_id = 1
            for corpus_type in self.switch_get:
                corpus_id = corpus_id + len(self.switch_get[corpus_type])
            annotable = True
            self.dictionary_corpus_list.append(
                DictionaryCorpus(
                    self.spreadsheet_id,
                    corpus_id,
                    f"{DictionaryCorpus.DEFAULT_SHEET_NAME}-{regional_settings_lemma}",
                    "",
                    self.tokenizer,
                    self.dictionary,
                    regional_settings_lemma,
                    annotable,
                    hash_lemma_dict,
                    corpus_dataframe=corpus_dataframe,
                )
            )
            self._update_corpus_table_by_id(
                corpus_id, self.dictionary_corpus_list[-1]
            )

    def create_synthetic_corpus(
        self,
        id,
        main_sheet,
        extension_sheet,
        variables_sheet,
        language_settings,
        json,
        annotable,
        ontology_instances,
        ontology_defaults,
        ontology_vertices,
        ontology_edges,
        ontology_properties,
    ):
        self.synthetic_corpus_list.append(
            SyntheticCorpus(
                self.spreadsheet_id,
                id,
                self.dictionary,
                self.standard,
                language_settings,
                json,
            )
        )
        self._update_corpus_table_by_id(id, self.synthetic_corpus_list[-1])

    def create_hugging_face_fixed_corpus(
        self,
        id,
        main_sheet,
        extension_sheet,
        variables_sheet,
        language_settings,
        json,
        annotable,
        ontology_instances,
        ontology_defaults,
        ontology_vertices,
        ontology_edges,
        ontology_properties,
    ):
        json_parsed = json_module.loads(json)
        self.hugging_face_corpus_list.append(
            HuggingFaceFixedCorpus(self.spreadsheet_id, id, json_parsed,)
        )
        self._update_corpus_table_by_id(id, self.hugging_face_corpus_list[-1])

    def create_mammut_corpus(
        self,
        id,
        main_sheet,
        extension_sheet,
        variables_sheet,
        language_settings,
        json,
        annotable,
        ontology_instances,
        ontology_defaults,
        ontology_vertices,
        ontology_edges,
        ontology_properties,
    ):
        if extension_sheet == CorpusMap.DEFAULT_NONE_KEY:
            extension_sheet = ""

        # TODO: Implementar la carga del nuevo Knowledge
        self.mammut_corpus_list.append(
            MammutCorpus(
                self.spreadsheet_id,
                id,
                main_sheet,
                extension_sheet,
                variables_sheet,
                self.tokenizer,
                self.dictionary,
                language_settings,
                annotable,
            )
        )
        self._update_corpus_table_by_id(id, self.mammut_corpus_list[-1])

    def create_crawler_corpus(
        self,
        id,
        main_sheet,
        extension_sheet,
        variables_sheet,
        language_settings,
        json,
        annotable,
        ontology_instances,
        ontology_defaults,
        ontology_vertices,
        ontology_edges,
        ontology_properties,
    ):
        if extension_sheet == CorpusMap.DEFAULT_NONE_KEY:
            extension_sheet = ""

        self.crawler_corpus_list.append(
            CrawlerCorpus(
                self.spreadsheet_id,
                id,
                main_sheet,
                extension_sheet,
                self.tokenizer,
                self.dictionary,
                language_settings,
                json,
                self.update_status_text_box,
                annotable,
            )
        )

        self._update_corpus_table_by_id(id, self.crawler_corpus_list[-1])

    def create_natural_corpus(
        self,
        id,
        main_sheet,
        extension_sheet,
        variables_sheet,
        language_settings,
        json,
        annotable,
        ontology_instances,
        ontology_defaults,
        ontology_vertices,
        ontology_edges,
        ontology_properties,
    ):
        if extension_sheet == CorpusMap.DEFAULT_NONE_KEY:
            extension_sheet = ""

        self.natural_corpus_list.append(
            NaturalCorpus(
                self.spreadsheet_id,
                id,
                main_sheet,
                extension_sheet,
                self.tokenizer,
                self.dictionary,
                language_settings,
                annotable,
            )
        )

        self._update_corpus_table_by_id(id, self.natural_corpus_list[-1])

    def create_images_corpus(
        self,
        id,
        main_sheet,
        extension_sheet,
        variables_sheet,
        language_settings,
        json,
        annotable,
        ontology_instances,
        ontology_defaults,
        ontology_vertices,
        ontology_edges,
        ontology_properties,
    ):
        if extension_sheet == CorpusMap.DEFAULT_NONE_KEY:
            extension_sheet = ""

        self.images_corpus_list.append(
            ImagesCorpus(
                self.spreadsheet_id,
                id,
                main_sheet,
                extension_sheet,
                self.tokenizer,
                self.dictionary,
                language_settings,
                self.update_status_text_box,
                annotable,
                json,
            )
        )

        self._update_corpus_table_by_id(id, self.images_corpus_list[-1])

    def create_issue_corpus(
        self,
        id,
        main_sheet,
        extension_sheet,
        variables_sheet,
        language_settings,
        json,
        annotable,
        ontology_instances,
        ontology_defaults,
        ontology_vertices,
        ontology_edges,
        ontology_properties,
    ):
        if extension_sheet == CorpusMap.DEFAULT_NONE_KEY:
            extension_sheet = ""
        knowledge = None
        self.issues_corpus_list.append(
            IssueCorpus(
                self.spreadsheet_id,
                id,
                main_sheet,
                extension_sheet,
                variables_sheet,
                self.tokenizer,
                self.dictionary,
                language_settings,
                annotable,
            )
        )

        self._update_corpus_table_by_id(id, self.issues_corpus_list[-1])

    def get_corpus(self, type, sheet_tittle):
        corpus_list = self.switch_get[type]
        for corpus in corpus_list:
            if corpus.sheet_title == sheet_tittle:
                return corpus
        raise NameError()

    def get_corpus_by_type(self, corpus_type, sheet_tittle):
        corpus_list = []
        for i in self.switch_get:
            if len(self.switch_get[i]) > 0:
                if type(self.switch_get[i][0]) == corpus_type:
                    corpus_list = self.switch_get[i]
        for corpus in corpus_list:
            if corpus.sheet_title == sheet_tittle:
                return corpus

    def get_all_corpus(self):
        return (
            self.mammut_corpus_list
            + self.natural_corpus_list
            + self.crawler_corpus_list
            + self.images_corpus_list
            + self.dictionary_corpus_list
            + self.issues_corpus_list
            + self.synthetic_corpus_list
            + self.hugging_face_corpus_list
        )

    def get_corpus_by_id(self, corpus_id: int):
        """Returns a corpus stored in the corpus map by id.

        Raises:
            KeyError if the corpus_id key isn't found in the corpus table

        Return:
            The corpus reference object.
        """
        return self._corpus_table_by_id[corpus_id]



from collections import namedtuple
import parsec as pc
from typing import List


class InflectionRule:
    """
    Supported commands:
        <E>: Empty string. DOes not modify the word
        <L>: keyboard Left arrow
        <R>: keyboard Right arrow
        <A>: remove accute accent from last letter aeiou
        <B>: keyboard Backspace
        <Á>: add acute accent to last letter aeiou
        <Ä>: add dieresis/trema/umlaut to last letter aeiou
    """

    VOWEL_REMOVE_ACUTE_ACCENT_REPLACEMENTS = {
        "á": "a",
        "é": "e",
        "í": "i",
        "ó": "o",
        "ú": "u",
    }
    VOWEL_ADD_ACUTE_ACCENT_REPLACEMENTS = {
        "a": "á",
        "e": "é",
        "i": "í",
        "o": "ó",
        "u": "ú",
    }
    VOWEL_ADD_DIERESIS_REPLACEMENTS = {
        "a": "ä",
        "e": "ë",
        "i": "ï",
        "o": "ö",
        "u": "ü",
    }
    PARADIGM_NAME_REGEX_STR = r"[A-ZÑÁÉÍÓÚÜÇa-zñáéíóúüç\-_0-9]+"

    Morpheme = namedtuple(
        "Morpheme",
        ["text", "features", "regional_settings", "deleted", "added"],
    )

    class PartialMorpheme:
        def __init__(self, current, cursor):
            self.current = current
            self.cursor = cursor
            self.deleted = []
            self.added = []

    def __init__(self, raw_rule: str, paradigm_producer):
        self.raw_rule = raw_rule
        self.paradigm_producer = paradigm_producer
        if "/" in raw_rule:
            rr_parts = raw_rule.split("/")
            self.commands = self.commands_parser().parse(rr_parts[0])
            self.features = []
            for f in rr_parts[1].split("+"):
                if (
                    f
                    in self.paradigm_producer.group.standard.feature_descriptor_list
                ):
                    self.features.append(
                        self.paradigm_producer.group.standard.feature_descriptor_list[
                            f
                        ]
                    )
                else:
                    raise Exception(
                        f"Regla de inflexion invalida. Id de feature invalido: \n{f} - {self.paradigm_producer.group.sheet_title} - {self.paradigm_producer.word_model} - {raw_rule}"
                    )
        else:
            raise Exception(
                f'Regla de inflexion invalida. El identificador es invalido o se omitio el separador "/". \n{self.paradigm_producer.group.sheet_title} - {self.paradigm_producer.word_model} - {raw_rule}'
            )

    def get_empty_command_parser(self):
        res = pc.string("<E>").result(self.empty_command)
        return res

    def empty_command(self, partial_morpheme: PartialMorpheme):
        return partial_morpheme

    def get_left_cursor_command_parser(self):
        res = pc.regex(r"<L[0-9]*>").parsecmap(
            lambda chars: self.create_left_cursor_command(chars)
        )
        return res

    def create_left_cursor_command(self, chars: str):
        chars = chars.strip("<L>")
        count = 1
        if chars:
            count = int(chars)

        def left_cursor_command(
            partial_morpheme: InflectionRule.PartialMorpheme,
        ):
            partial_morpheme.cursor -= count
            return partial_morpheme

        return left_cursor_command

    def get_right_cursor_command_parser(self):
        res = pc.regex(r"<R[0-9]*>").parsecmap(
            lambda chars: self.create_right_cursor_command(chars)
        )
        return res

    def create_right_cursor_command(self, chars: str):
        chars = chars.strip("<R>")
        count = 1
        if chars:
            count = int(chars)

        def right_cursor_command(
            partial_morpheme: InflectionRule.PartialMorpheme,
        ):
            partial_morpheme.cursor += count
            return partial_morpheme

        return right_cursor_command

    def get_remove_acute_accent_command_parser(self):
        res = pc.string("<A>").result(self.remove_acute_accent_command)
        return res

    def remove_acute_accent_command(self, partial_morpheme: PartialMorpheme):
        vowel_accent = partial_morpheme.current[partial_morpheme.cursor]
        temp_list = list(partial_morpheme.current)
        if (
            vowel_accent
            in InflectionRule.VOWEL_REMOVE_ACUTE_ACCENT_REPLACEMENTS
        ):
            temp_list[
                partial_morpheme.cursor
            ] = InflectionRule.VOWEL_REMOVE_ACUTE_ACCENT_REPLACEMENTS[
                vowel_accent
            ]
        else:
            raise Exception("El caracter no es una vocal acentuada.")
        partial_morpheme.current = "".join(temp_list)
        return partial_morpheme

    def get_add_acute_accent_command_parser(self):
        res = pc.string("<Á>").result(self.add_acute_accent_command)
        return res

    def add_acute_accent_command(self, partial_morpheme: PartialMorpheme):
        vowel_accent = partial_morpheme.current[partial_morpheme.cursor]
        temp_list = list(partial_morpheme.current)
        if vowel_accent in InflectionRule.VOWEL_ADD_ACUTE_ACCENT_REPLACEMENTS:
            temp_list[
                partial_morpheme.cursor
            ] = InflectionRule.VOWEL_ADD_ACUTE_ACCENT_REPLACEMENTS[
                vowel_accent
            ]
        else:
            raise Exception("El caracter no es una vocal sin acento.")
        partial_morpheme.current = "".join(temp_list)
        return partial_morpheme

    def get_add_dieresis_command_parser(self):
        res = pc.string("<Ä>").result(self.add_dieresis_command)
        return res

    def add_dieresis_command(self, partial_morpheme: PartialMorpheme):
        vowel_dieresis = partial_morpheme.current[partial_morpheme.cursor]
        temp_list = list(partial_morpheme.current)
        if vowel_dieresis in InflectionRule.VOWEL_ADD_DIERESIS_REPLACEMENTS:
            temp_list[
                partial_morpheme.cursor
            ] = InflectionRule.VOWEL_ADD_DIERESIS_REPLACEMENTS[vowel_dieresis]
        else:
            raise Exception("El caracter no es una vocal sin dieresis.")
        partial_morpheme.current = "".join(temp_list)
        return partial_morpheme

    def get_delete_current_char_command_parser(self):
        res = pc.regex(r"<B[0-9]*>").parsecmap(
            lambda chars: self.create_delete_current_char_command(chars)
        )
        return res

    def create_delete_current_char_command(self, chars: str):
        chars = chars.strip("<B>")
        count = 1
        if chars:
            count = int(chars)

        def delete_current_char_command(
            partial_morpheme: InflectionRule.PartialMorpheme,
        ):
            temp_list = list(partial_morpheme.current)
            for i in range(0, count):
                removed_char = temp_list.pop(partial_morpheme.cursor)
                partial_morpheme.cursor -= 1
                partial_morpheme.deleted.append(removed_char)
            partial_morpheme.current = "".join(temp_list)
            return partial_morpheme

        return delete_current_char_command

    def get_insert_in_current_position_command_parser(self):
        res = pc.regex(InflectionRule.PARADIGM_NAME_REGEX_STR).parsecmap(
            lambda chars: self.create_insert_in_current_position_command(chars)
        )
        return res

    def create_insert_in_current_position_command(self, chars: str):
        chars = chars.lower()

        def insert_in_current_position_command(
            partial_morpheme: InflectionRule.PartialMorpheme,
        ):
            temp_list = list(partial_morpheme.current)
            for c in chars:
                partial_morpheme.cursor += 1
                temp_list.insert(partial_morpheme.cursor, c)
                partial_morpheme.added.append(c)
            partial_morpheme.current = "".join(temp_list)
            return partial_morpheme

        return insert_in_current_position_command

    def commands_parser(self):
        options = (
            self.get_left_cursor_command_parser()
            ^ self.get_right_cursor_command_parser()
            ^ self.get_add_acute_accent_command_parser()
            ^ self.get_add_dieresis_command_parser()
        )
        options = (
            options
            ^ self.get_remove_acute_accent_command_parser()
            ^ self.get_delete_current_char_command_parser()
        )
        options = (
            options
            ^ self.get_empty_command_parser()
            ^ self.get_insert_in_current_position_command_parser()
        )
        return pc.many(options)

    def apply(self, lemma: str) -> Morpheme:
        partialMorpheme = InflectionRule.PartialMorpheme(lemma, len(lemma) - 1)
        for c in self.commands:
            partialMorpheme = c(partialMorpheme)
        return InflectionRule.Morpheme(
            partialMorpheme.current,
            self.features,
            self.paradigm_producer.group.regional_setting,
            partialMorpheme.deleted,
            partialMorpheme.added,
        )



# coding=utf-8
from ipywidgets import *
from IPython.display import display
from django.template.loader import get_template
import chart_studio.plotly as py
import plotly.graph_objs as go

invalid_spans_table_template = get_template("invalid_spans_table.html")


def get_validator_widget(package):
    def event_validator_selector(corpus):
        # print('event validator selector')
        all_ea_validator = package.corpus_annotations_validator.annotations_validator_by_corpus[
            corpus.sheet_title
        ]
        all_events_dict = {}
        if valid_checkbox.value:
            events = all_ea_validator.valid_events
        else:
            events = all_ea_validator.invalid_events

        for annotaion_event_validator in events:
            for invalid_type in annotaion_event_validator.invalid_type:
                all_events_dict[
                    annotaion_event_validator.annotation.event["original_text"]
                    + invalid_type
                ] = annotaion_event_validator
        all_events_dict["None"] = "None"
        events_dropdown = widgets.Dropdown(
            options=all_events_dict,
            value="None",
            description="all events:",
            layout=Layout(width="800px"),
        )

        def show_overlaps(new_validator_event, selected_dimension):
            visualizers_list = all_ea_validator.visualize_overlap_from_event_validator(
                new_validator_event, selected_dimension
            )
            # TODO: de aqui en adelante es util el codigo o era cuando mostraba la grafica??
            # overlap_pair_type, events_appearence = all_ea_validator.get_event_overlap_and_overall_count(
            #    new_validator_event)
            # if overlap_pair_type is not None:
            #    overlap_count = all_ea_validator.get_event_overlap_and_overall_count(new_validator_event)
            return visualizers_list

        def dimension_selector(
            new_validator_event,
        ):  # TODO:  poner cual dimension se esta editando??
            event_annotation = new_validator_event.annotation
            # print('dimesion selector', event_annotation)
            dimensions_dict = {}
            for dimension in event_annotation.dimensions:
                dimensions_dict[dimension] = dimension
            dimensions_dict["None"] = "None"

            dimensions_dropdown = widgets.Dropdown(
                options=dimensions_dict,
                value="None",
                description="select dimension:",
                layout=Layout(width="500px"),
            )

            def dimensions_dropdown_change(change):
                dimension = change["new"]
                if not only_overlaps_checkbox.value:
                    selected_dimension_visualizer = event_annotation.get_visualizer(
                        dimension, None
                    )
                    visualizer_list = show_overlaps(new_validator_event, dimension)
                    dimension_dropdown_vbox.children = [
                        dimensions_dropdown,
                        selected_dimension_visualizer,
                    ] + visualizer_list
                else:
                    visualizer_list = show_overlaps(new_validator_event, dimension)
                    if len(visualizer_list) == 0:
                        no_overlap_message = widgets.HTML(
                            value=dimension + " no tiene error de solapado"
                        )
                        dimension_dropdown_vbox.children = [
                            dimensions_dropdown,
                            no_overlap_message,
                        ]
                    else:
                        dimension_dropdown_vbox.children = [
                            dimensions_dropdown
                        ] + visualizer_list

            dimensions_dropdown.observe(dimensions_dropdown_change, names="value")
            dimension_dropdown_vbox = widgets.VBox([dimensions_dropdown])
            return dimension_dropdown_vbox

        def events_dropdown_change(change):
            if change["new"] is not "None":
                new_validator_event = change["new"]  # TODO: mejorar este nombre
                # print('new validator envent', new_validator_event)
                dimension_selector_widget = dimension_selector(new_validator_event)
                events_dropdow_selector_section.children = [
                    events_dropdown,
                    dimension_selector_widget,
                ]
            else:
                print("no se selecciono corpus")

        events_dropdown.observe(events_dropdown_change, names="value")

        show_chart_button = widgets.Button(
            description="mostrar charts",
            button_style="",
            tooltip="mostrar charts",
            icon="",
            layout=Layout(width="200px"),
        )

        def handle_show_chart_click(sender):
            new_validator_event = events_dropdown.value
            # display(all_ea_validator.visualize_overlap_from_event_validator(new_validator_event))
            (
                overlap_pair_type,
                events_appearence,
            ) = all_ea_validator.get_event_overlap_and_overall_count(
                new_validator_event
            )
            if overlap_pair_type is not None:
                overlap_count = all_ea_validator.get_event_overlap_and_overall_count(
                    new_validator_event
                )
                # print(overlap_count)
                # fig = tools.make_subplots(rows=2, cols=1)
                data = []
                annotations = []
                for i, vals in enumerate(overlap_count[1]):
                    # print(overlap_count[1])
                    labels = ["valid" + str(i), "invalid" + str(i)]
                    values = [overlap_count[1][vals], 1]
                    # print(labels, values)

                    if i == 0:
                        domain = {"x": [0, 0.5]}
                        text_domain = {"x": [0, 0.5], "y": [0, 0.5]}

                    else:
                        domain = {"x": [0.51, 1]}
                        text_domain = {"x": [0.51, 1], "y": [0, 0.5]}

                    trace = {
                        "labels": labels,
                        "values": values,
                        "type": "pie",
                        "name": list(overlap_count[1].keys())[i],
                        "domain": domain,
                        "hoverinfo": "label+percent+name",
                        "textinfo": "none",
                    }
                    data.append(trace)
                    annotation = {
                        "font": {"size": 20},
                        "showarrow": False,
                        "text": list(overlap_count[1].keys())[i],
                        "x": text_domain["x"][0],
                        "y": text_domain["y"][0],
                    }
                    annotations.append(annotation)
                layout = {
                    "title": "porcentaje del evento solapado conrespecto al total de cada anotacion",
                    "annotations": [],
                }
                fig = dict()
                fig["data"] = data
                fig["layout"] = layout
                fig["layout"]["annotations"] = annotations

                display(py.iplot(fig, filename="donut"))

        show_chart_button.on_click(handle_show_chart_click)
        events_dropdow_selector_section = widgets.VBox([events_dropdown])
        return events_dropdow_selector_section

    all_corpus = package.corpus_map.get_all_corpus()
    all_corpus_dict = {}
    for i in all_corpus:
        all_corpus_dict[i.sheet_title] = i
    all_corpus_dict["None"] = "None"
    corpus_dropdown = widgets.Dropdown(
        options=all_corpus_dict,
        value="None",
        description="select corpus:",
        layout=Layout(width="500px"),
    )

    valid_checkbox = widgets.Checkbox(
        value=False, description="eventos validos", disabled=False
    )

    only_overlaps_checkbox = widgets.Checkbox(
        value=False, description="solo solapamiento", disabled=False
    )
    checkboxes = widgets.HBox([valid_checkbox, only_overlaps_checkbox])

    def valid_checkbox_change(d):
        corpus_dropdown.value = "None"
        validator_widget.children = [checkboxes, corpus_dropdown]

    valid_checkbox.observe(valid_checkbox_change)
    only_overlaps_checkbox.observe(valid_checkbox_change)

    def corpus_dropdown_change(change):
        if change["new"] != "None":
            validator_widget.children = [
                checkboxes,
                corpus_dropdown,
                event_validator_selector(change["new"]),
            ]
        else:
            validator_widget.children = [checkboxes, corpus_dropdown]

    corpus_dropdown.observe(corpus_dropdown_change, names="value")

    validator_widget = widgets.VBox([checkboxes, corpus_dropdown])
    return validator_widget


def get_invalid_spans_list(package):
    def display_invalid_spans(events_validator):
        definitions = [
            {"overlap": i[0], "original_text": i[1], "dimension": i[2]}
            for i in events_validator.overlap_spans_list
        ]
        context_dict = {"spans_values": definitions}
        invalid_spans_table_html = invalid_spans_table_template.render(context_dict)
        invalid_spans = widgets.HTML(value=invalid_spans_table_html)
        return invalid_spans

    all_corpus = package.corpus_map.get_all_corpus()
    all_corpus_dict = {}
    for i in all_corpus:
        all_corpus_dict[i.sheet_title] = i
    all_corpus_dict["None"] = "None"
    corpus_dropdown = widgets.Dropdown(
        options=all_corpus_dict,
        value="None",
        description="select corpus:",
        layout=Layout(width="500px"),
    )

    def corpus_dropdown_change(change):
        if change["new"] != "None":
            corpus = change["new"]
            all_ea_validator = package.corpus_annotations_validator.annotations_validator_by_corpus[
                corpus.sheet_title
            ]
            validator_widget.children = [
                corpus_dropdown,
                display_invalid_spans(all_ea_validator),
            ]
        else:
            validator_widget.children = [corpus_dropdown]

    corpus_dropdown.observe(corpus_dropdown_change, names="value")

    validator_widget = widgets.VBox([corpus_dropdown])
    return validator_widget


def get_valid_and_invalid_events_graph(all_ea_validator):
    x = [i.valid for i in all_ea_validator.event_annotation_validator_list]
    data = [go.Histogram(x=x)]
    return py.iplot(data, filename="basic histogram")


def get_overlaps_by_dimension_graph(all_ea_validator):
    event_dim_keys = list()
    event_dim_values = list()
    entities_dim_keys = list()
    entities_dim_values = list()
    for key, val in all_ea_validator.invalid_by_dimension.items():
        for list_type, dimension_val in val.items():
            if list_type == "events_list":
                event_dim_keys.append(key)
                event_dim_values.append(len(dimension_val))
            else:
                entities_dim_keys.append(key)
                entities_dim_values.append(len(dimension_val))
    events_bars = go.Bar(x=event_dim_keys, y=event_dim_values, name="invalid events")
    entities_bars = go.Bar(
        x=entities_dim_keys, y=entities_dim_values, name="invalid entities"
    )
    data = [events_bars, entities_bars]

    layout = go.Layout(barmode="group")

    fig = go.Figure(data=data, layout=layout)
    return py.iplot(fig, filename="grouped-bar")


def get_overlaps_by_annotation_pair(all_ea_validator):
    keys = list()
    values = list()
    for key, val in all_ea_validator.invalid_annotations_pairs_count.items():
        keys.append(key)
        values.append(val)
    data = [go.Bar(x=keys, y=values)]

    return py.iplot(data, filename="basic-bar")


def get_overlaps_heatmap_by_annotation_pair(all_ea_validator):
    keys = list()
    values = list()
    for key, val in all_ea_validator.invalid_annotations_pairs_count.items():
        keys.append(key)
        values.append(val)

    x = [i.split("-")[0] for i in keys]
    y = [i.split("-")[1] for i in keys]
    trace = go.Heatmap(z=values, x=x, y=y)
    data = [trace]
    return py.iplot(data, filename="labelled-heatmap")


def get_overlap_representation_vs_appearance(all_ea_validator):
    pair_annotation_count_dict = dict()
    for invalid_event in all_ea_validator.invalid_events:
        overlap_pair_type_count = all_ea_validator.get_event_overlap_and_overall_count(
            invalid_event
        )
        for pair in overlap_pair_type_count:
            pair_annotation_count_dict[pair] = overlap_pair_type_count[pair]
    # debemos sacar los que no estan en parir_annotation_count_dict
    clean_annotation_pairs_count = dict()
    for key, val in all_ea_validator.invalid_annotations_pairs_count.items():
        if key in pair_annotation_count_dict:
            clean_annotation_pairs_count[key] = val
        else:
            print("key not in dict", key, "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<")
    data_list = list()
    overlap_data_list = list()
    x = list()
    index = 0
    for key, val in clean_annotation_pairs_count.items():
        index += 1
        print(key, val)
        if index % 2 == 0:
            total_count_color = "#448"
        else:
            total_count_color = "#458917"
        if key in pair_annotation_count_dict:
            print(pair_annotation_count_dict[key])
            # veces que aparecen el evento
            a1 = go.Bar(
                x=list(pair_annotation_count_dict[key].keys()),
                y=list(pair_annotation_count_dict[key].values()),
                type="bar",
                name=key,
                xaxis="x" + str(index),
                marker={"color": total_count_color},
            )
            data_list.append(a1)
            # veces que se solapa
            a2 = go.Bar(
                x=list(pair_annotation_count_dict[key].keys()),
                y=[val, val],
                type="bar",
                name=key + " solapados",
                xaxis="x" + str(index),
                marker={"color": "#CCF"},
            )
            overlap_data_list.append(a2)

        else:
            print("key not in dict", key, "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<")
            # TODO: porque hay un key que no coincide????
        # print(key, parir_annotation_count_dict[key], val)

    # creamos el alyout para agrupar los por eventos
    layout_vals = dict(
        barmode="stack",
        # yaxis= {'tickformat': '%'}
    )
    data_len = len(data_list)
    step_size = 1 / data_len
    step_val = 0
    for i, list_elem in enumerate(data_list):
        if i == 0:
            layout_vals["xaxis"] = dict(
                domain=[step_val, step_val + step_size], anchor=list_elem["xaxis"],
            )
        else:
            layout_vals["xaxis" + str(i + 1)] = dict(
                domain=[step_val, step_val + step_size], anchor=list_elem["xaxis"],
            )
        step_val = step_val + step_size
    # unimos las dos listas(cuanta de solapados y aparciones de eventoss)
    graph_data = list()
    for index, i in enumerate(data_list):
        graph_data.append(overlap_data_list[index])
        graph_data.append(i)

    data = graph_data

    fig = go.Figure(data=data, layout=layout_vals)
    return py.iplot(fig, filename="")



from mammut.common.basic_tokenizer import BasicTokenizer
from mammut.common.corpus.constants import CorpusTable, CrawlerCorpusTable
from mammut.common.corpus.corpus_base import Corpus, CorpusScenarioEntry
from mammut.common.lexicon.dictionary import Dictionary
from mammut.common.storage.storage_manager import StorageManager
from mammut.data.mammut_crawler import CrawlingProperties, MammutCrawler


class CrawlerCorpus(Corpus):
    storage_manager = StorageManager()

    def __init__(
            self,
            spreadsheet_id: str,
            id,
            corpus_sheet_title: str,
            corpus_extension_sheet_title: str,
            tokenizer: BasicTokenizer,
            dictionary: Dictionary,
            regional_settings: str,
            json_parameters,
            update_status_text_box,
            annotable: bool,
    ):
        self.sheet_title = corpus_sheet_title
        self.spreadsheet_id = spreadsheet_id
        self.update_status_text_box = update_status_text_box
        Corpus.__init__(
            self,
            spreadsheet_id,
            id,
            corpus_sheet_title,
            corpus_extension_sheet_title,
            tokenizer,
            dictionary,
            regional_settings,
            annotable,
            json_parameters,
        )
        self.corpus_sheet_title = corpus_sheet_title

    def create_dump_sheet(self, spreadsheet_id):
        temporal_sheet_title = self.storage_manager.create_temporal_new_sheet(
            spreadsheet_id, self.sheet_title
        )
        self.storage_manager.add_rows_to_spreadsheet(
            spreadsheet_id,
            temporal_sheet_title,
            CrawlerCorpusTable.DEFAULT_COLUMN_TITTLES,
            CrawlerCorpusTable.DEFAULT_START_ROW_TITTLE,
            CrawlerCorpusTable.DEFAULT_START_COLUMN_TITTLE,
        )
        return temporal_sheet_title

    def get_corpus_scenario_entry_from_row(self, g):
        return CorpusScenarioEntry(
            g[0],
            g[1].iloc[0][CorpusTable.SCENARIO_TYPE_COLUMN_NAME()],
            g[1],
            self.tokenizer,
            self.dictionary,
            self.regional_settings,
            self,
        )

    def load_data(self):
        data = []
        i = 0
        crawling_properties = CrawlingProperties(
            self.configuration[CrawlerCorpusTable.URL_KEY],
            self.configuration[CrawlerCorpusTable.NODES_PATTERN_KEY],
            self.configuration[CrawlerCorpusTable.NODE_ELEMENT_KEY],
            self.configuration[CrawlerCorpusTable.PAGE_HREF_KEY],
            pagination_nodes_pattern=self.configuration[
                CrawlerCorpusTable.PAGINATION_NODE_PATTERN_KEY
            ],
            pagination_element_xpath=self.configuration[
                CrawlerCorpusTable.PAGINATION_NODE_ELEMENT_KEY
            ],
            pages_limit=self.configuration[
                CrawlerCorpusTable.DEFAULT_PAGES_LIMIT_KEY
            ],
        )
        crawling_data = MammutCrawler(
            crawling_properties,
            update_status_text_box=self.update_status_text_box,
        ).execute()
        for dict in crawling_data:
            i += 1
            data.append(
                [
                    i,
                    i,
                    CrawlerCorpusTable.DEFAULT_SCENARIO_TYPE_VALUE,
                    dict["element"],
                    CrawlerCorpusTable.DEFAULT_EVENT_MESSAGE_VALUE,
                    CrawlerCorpusTable.DEFAULT_HIDDEN_VALUE,
                    CrawlerCorpusTable.DEFAULT_FIELD_VALUE,
                    CrawlerCorpusTable.DEFAULT_SOURCE_VALUE,
                    self.regional_settings,
                    CrawlerCorpusTable.DEFAULT_COMPLEXITY_VALUE,
                ]
            )
        self.write_data(self.DEFAULT_CORPUS_RANGE, data)

    def write_data(self, range, rows):
        sheet_name, range_cell = range.split("!")
        self.storage_manager.add_rows_to_spreadsheet(
            self.spreadsheet_id,
            sheet_name,
            rows,
            range_cell[1: len(range_cell)],
            range_cell[0],
        )




from dynaconf import Dynaconf
import os

current_directory = os.path.dirname(os.path.realpath(__file__))

settings = Dynaconf(
    envvar_prefix="DYNACONF",
    settings_files=[f"{current_directory}/settings.toml", f"{current_directory}/.secrets.toml"],
)

HOME = os.path.expanduser("~")
save_folder = os.path.join(HOME, '.mammut-py', 'curriculum')

# `envvar_prefix` = export envvars with `export DYNACONF_FOO=bar`.
# `settings_files` = Load this files in the order.




from mammut.curriculum.models import model_base
import mammut.curriculum.report as report
import time
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class TestMockModel(model_base.ModelBase):
    def __init__(self, course_id: int, lesson_id: int):
        model_base.ModelBase.__init__(self, {}, course_id, lesson_id)
        self.sleep_time = 20

    def train(self, mammut_session_context: MammutSessionContext, **kwargs):
        time.sleep(self.sleep_time)

    def save(self, mammut_session_context: MammutSessionContext, **kwargs):
        model_file_system_path = self._get_model_file_system_path(
            mammut_session_context.curriculum_save_folder
        )
        report.send_message(
            report.CurriculumGeneralDebugMessage(
                f"Save TestMockModel model in: {model_file_system_path}"
            )
        )
        pass



from mammut.common.storage.storage_manager import StorageManager
from mammut.assessment.assessor import Assessor
from mammut.curriculum.course.course import Course
from mammut.curriculum.report.package_progress_report import PackageProgress
import mammut.curriculum.course.utils as course_utils
from mammut.curriculum.core.flowchart import Flowchart
from mammut.curriculum.lesson import lesson_base
from mammut.curriculum.lesson.lesson_type import LessonType
from mammut.common.package import Package
from typing import List, Optional
import pandas as pd
import mammut.curriculum.report as report
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class CurriculumBase(object):
    """Encapsulates all the Curriculum learning required tools:
        - Courses, Lessons, Assessments.
        - Package, Dictionary, LinguisticStandard.

    The courses in a Curriculum forms a flowchart, which is
    a directed acyclic graph. The flowchart implies that courses are
    meant to be progressive, which allows to organize them into academic
    periods.
    """

    _storage_manager = StorageManager()

    DEFAULT_START_CELL = "A1"
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "1"
    DEFAULT_END_CELL = "J"

    def __init__(
        self,
        package_id: str,
        standard_spreadsheet_id: str,
        curriculum_base_sheet_title: str = "curriculum",
        lessons_base_sheet_title: str = "lessons",
        assessments_sheet_title: str = "assessments",
        standard_base_sheet_title: str = "base",
        standard_dict_sources_sheet_title: str = "dict-sources",
        standard_pos_sheet_title: str = "pos-tags",
        standard_features_sheet_title: str = "features",
        standard_dimensions_sheet_title: str = "dimensions",
        dictionary_sheet_title: str = "dictionary",
        standard_functionals_sheet_title: str = "functionals",
        standard_non_functionals_sheet_title: str = "non-functionals",
        standard_primes_sheet_title: str = "primes",
        standard_synonym_sheet_title: str = "synonym",
        synonym_sheet_title: str = "synonym",
        progress_sheet_title: str = "progress",
        package_id_is_spreadsheet: bool = False,
        show_progress: bool = True,
        create_progress_sheet: bool = False,
        mammut_session_context: Optional[MammutSessionContext] = None,
        enable_ktt_interpretation: bool = False,
    ):
        self.errors = []
        self.valid = True
        self.curriculum_base_sheet_title: str = curriculum_base_sheet_title
        self.lessons_base_sheet_title: str = lessons_base_sheet_title
        self.assessment_sheet_title: str = assessments_sheet_title
        self.progress_sheet_title: str = progress_sheet_title
        self._create_progress_sheet: bool = create_progress_sheet
        self.package: Package = Package(
            package_id,
            standard_spreadsheet_id,
            package_id_is_spreadsheet,
            show_progress,
            standard_base_sheet_title,
            standard_dict_sources_sheet_title,
            standard_pos_sheet_title,
            standard_features_sheet_title,
            standard_dimensions_sheet_title,
            dictionary_sheet_title,
            standard_functionals_sheet_title,
            standard_non_functionals_sheet_title,
            standard_primes_sheet_title,
            standard_synonym_sheet_title,
            synonym_sheet_title,
        )
        self.package_spreadsheet_id = self.package.spreadsheet_id
        self.courses: List[Course] = []
        self.flowchart: Optional[Flowchart] = None
        self.assessor: Optional[Assessor] = None
        self.base_curriculum_dataframe = None
        self.base_lessons_dataframe = None
        self.progress: Optional[PackageProgress] = None
        self._enable_ktt_interpretation: bool = enable_ktt_interpretation

        # TODO: The session context is created here if doesn't exists,
        #   because right now, the Curriculum is the entrypoint class.
        #   In the future, there'll be another layer to instantiate Curriculum processes.
        #   Sessions should be created there.
        if mammut_session_context:
            self._mammut_session_context = mammut_session_context
        else:
            self._mammut_session_context = MammutSessionContext(
                self.package.main_id,
            )
            report.set_mammut_session_context(self._mammut_session_context)

        super(CurriculumBase, self).__init__()

    def _load_curriculum_spreadsheets(self):
        """Loads information from curriculum sheets.

        Downloads the dataframes for:
            - curriculum spreadsheet
            - lessons spreadsheet
        """
        base_curriculum_dataframe: pd.DataFrame = self._storage_manager.get_spreadsheet_as_dataframe(
            self.package_spreadsheet_id,
            self.curriculum_base_sheet_title,
            self.DEFAULT_START_COLUMN,
            self.DEFAULT_START_ROW,
            self.DEFAULT_END_CELL,
        )
        if base_curriculum_dataframe is None:
            self.valid = False
            message = f"{self.curriculum_base_sheet_title} sheet not found"
            self.errors += message
            report.send_message(report.CurriculumGeneralErrorMessage(message))

        base_lessons_dataframe = self._storage_manager.get_spreadsheet_as_dataframe(
            self.package_spreadsheet_id,
            self.lessons_base_sheet_title,
            lesson_base.Lesson.DEFAULT_START_COLUMN,
            lesson_base.Lesson.DEFAULT_START_ROW,
            lesson_base.Lesson.DEFAULT_END_CELL,
        )
        if base_lessons_dataframe is None:
            self.valid = False
            message = f"{self.lessons_base_sheet_title} sheet not found"
            self.errors += message
            report.send_message(report.CurriculumGeneralErrorMessage(message))

        """
            Validates the lessons type column values for specified courses
        """
        if not all(
            base_lessons_dataframe[lesson_base.Lesson.MODEL_COLUMN].isin(
                LessonType.names()
            )
        ):
            self.valid = False
            message = "Invalid Model(s) specified for Lessons"
            self.errors += message
            report.send_message(report.CurriculumGeneralErrorMessage(message))

        if self.valid:
            self.base_curriculum_dataframe = base_curriculum_dataframe
            self.base_lessons_dataframe = base_lessons_dataframe

            report.send_message(
                report.CurriculumGeneralDebugMessage(
                    "Load Courses from Package."
                )
            )

            for index, row_course in base_curriculum_dataframe.iterrows():
                if not row_course[course_utils.HIDDEN_COLUMN]:
                    course = self._get_course_instance(
                        row_course, base_lessons_dataframe
                    )
                    self.courses.append(course)

    def load(self):
        """Loads the Curriculum and all the Package information.
        """
        report.send_message(
            report.CurriculumGeneralInfoMessage(
                "Load Curriculum Package information",
            )
        )

        self.package.load_main_classes()

        report.send_message(
            report.CurriculumGeneralInfoMessage(
                "Main classes Package, LinguisticStandard, Dictionary loaded"
            )
        )

        """
        Assessor must load before the courses, since the courses
        instances contains a reference to the assessor.
        """
        self.assessor = Assessor(
            self.package_spreadsheet_id, self.assessment_sheet_title
        )

        self.assessor.load()

        report.send_message(
            report.CurriculumGeneralInfoMessage("Curriculum Assessor loaded")
        )

        self.progress = PackageProgress(
            self,
            self._storage_manager,
            self.progress_sheet_title,
            create_sheet=self._create_progress_sheet,
        )

        self._load_curriculum_spreadsheets()
        self.flowchart = Flowchart(self)
        if not self.flowchart.valid or not self.valid:
            self.valid = False
            self.errors += self.flowchart.errors

        if self._enable_ktt_interpretation:
            self.flowchart.add_ktt_interpretation_course(self)

        report.send_message(
            report.CurriculumGeneralInfoMessage(
                "Curriculum courses flowchart loaded"
            )
        )

    def _get_course_instance(
        self, row: pd.Series, lessons_df: pd.DataFrame
    ) -> Course:
        """Returns the concrete instantiation of a course.
        It performs concrete instantiation of lessons too
        that are associated to that course.

        Private method used at initialization time.
        """
        course = Course(
            row[course_utils.ID_COLUMN],
            row[course_utils.ACADEMIC_PERIOD_COLUMN],
            row[course_utils.PREREQUISITE_COLUMN],
            row[course_utils.LANGUAGE_LEVEL_COLUMN],
            row[course_utils.COURSE_NAME_COLUMN],
            row[course_utils.APPROVAL_CRITERIA_COLUMN],
            row[course_utils.RETRY_COLUMN],
            row[course_utils.REGIONAL_SETTINGS_COLUMN],
            Course.get_lessons_for_course(
                row[course_utils.ID_COLUMN],
                lessons_df,
                self.package.corpus_map,
                self._mammut_session_context,
            ),
            self.assessor,
            self._mammut_session_context,
        )
        return course



"""
    Course utilities constants
"""
ID_COLUMN = "id"
ACADEMIC_PERIOD_COLUMN = "academic_period"
PREREQUISITE_COLUMN = "prerequisites"
LANGUAGE_LEVEL_COLUMN = "language_level"
COURSE_NAME_COLUMN = "course_name"
REGIONAL_SETTINGS_COLUMN = "regional_settings"
HIDDEN_COLUMN = "hidden"
OBSERVATIONS_COLUMN = "observations"
APPROVAL_CRITERIA_COLUMN = "approval_criteria"
RETRY_COLUMN = "retry"



class BaseError(Exception):
    """
        Base class for curriculum processing exceptions.
    """

    def __init__(self, message: str, phase: str):
        self.message = message
        self.phase = phase

    def __str__(self):
        s = f"Error: {self.message}, Phase: {self.phase}"
        return s



from mammut.common.lexicon.linguistic_standard import LinguisticStandard
from typing import List
from mammut.common.synthetic.synthetic_corpus_generator.token_generator import TokenGenerator


class FunctionalGenerator(TokenGenerator):
    """Functional words generator Encapsulates the retrieval of
    functional words in the Standard.

    Only 'es' functional words are being retrieved, since only 'es'
    functional words are being loaded in the LinguisticStandard
    class.
    """

    POS_PARADIGM_COL_NAME = "pos-paradigm"
    LEMMA_COL_NAME = "lemma"

    def __init__(self, standard: LinguisticStandard):
        self.standard = standard
        self.func_df = standard.functionals_data_frame

    # Override abstract method
    def get_tokens(
        self, pos: str = "", regional_settings: str = "es"
    ) -> List[str]:
        if (
            pos != "PREP"
            and pos != "CONJ"
            and pos != "PRON"
            and pos != "AUX"
            and pos != "ART"
        ):
            return []

        func_pos = self.func_df[self.POS_PARADIGM_COL_NAME] == pos
        selected_df = self.func_df[func_pos]
        selected_df_values = set(selected_df[self.LEMMA_COL_NAME].values)
        if pos == "ART":
            selected_df_values.remove("zero-ES")
        ret = list(selected_df_values)
        return ret



# coding=utf-8
import mammut.common.google_api as ga
import os
import configparser
from graphviz import *
from IPython.display import *
from mammut.common.storage.storage_manager import StorageManager


class MammutGraph:

    storage_manager = StorageManager()
    DEFAULT_VERTEX_SHEET_TITTLE = "vertices"
    DEFAULT_VERTEX_BEGIN_RANGE = "A2"
    DEFAULT_VERTEX_BEGIN_COLUMN = "A"
    DEFAULT_VERTEX_BEGIN_ROW = "2"
    DEFAULT_VERTEX_END_RANGE = "G"

    DEFAULT_PROPERTIES_SHEET_TITTLE = "properties"
    DEFAULT_PROPERTIES_BEGIN_RANGE = "A2"
    DEFAULT_PROPERTIES_BEGIN_COLUMN = "A"
    DEFAULT_PROPERTIES_BEGIN_ROW = "2"
    DEFAULT_PROPERTIES_END_RANGE = "F"

    VERTEX_NAME = "name"
    VERTEX_ELEMENT_TYPE = "element_type"
    VERTEX_PROPERTY_PREDICATE = "property_predicate"
    VERTEX_ELEMENT_NAME = "element_name"
    VERTEX_EDGE_HEAD = "edge_head"
    VERTEX_OPTIONAL = "optional"
    VERTEX_ENTRY_POINT = "entry_point"

    PROPERTY_NAME = "name"
    PROPERTY_PROPERTY_TYPE = "property_type"
    PROPERTY_CARDINALITY = "cardinality"
    PROPERTY_PROPERTIES = "properties"
    PROPERTY_RANGE = "range"
    PROPERTY_LAMBDA_LEXICAL_FORM = "lambda_lexical_form"

    def __init__(self, spreadsheet_id, package_name):
        self.package_name = package_name
        self.vertex_data_frame = None
        self.vertex_group_by_object = None
        self.properties_data_frame = None
        self.properties_dict = {}
        self.graph = Graph("G", filename="MammutGraph.gv", format="png")
        self.load_graph(spreadsheet_id)

    def load_graph(self, spreadsheet_id):
        self.vertex_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            spreadsheet_id,
            MammutGraph.DEFAULT_VERTEX_SHEET_TITTLE,
            MammutGraph.DEFAULT_VERTEX_BEGIN_COLUMN,
            MammutGraph.DEFAULT_VERTEX_BEGIN_ROW,
            MammutGraph.DEFAULT_VERTEX_END_RANGE,
        )

        self.properties_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            spreadsheet_id,
            MammutGraph.DEFAULT_PROPERTIES_SHEET_TITTLE,
            MammutGraph.DEFAULT_PROPERTIES_BEGIN_COLUMN,
            MammutGraph.DEFAULT_PROPERTIES_BEGIN_ROW,
            MammutGraph.DEFAULT_PROPERTIES_END_RANGE,
        )

        self.graph.attr("node", shape="doublecircle")
        for vertex in self.vertex_data_frame.groupby([MammutGraph.VERTEX_NAME]):
            self.graph.node(vertex[0], vertex[0])

        self.graph.attr("node", style="filled", shape="square")
        for i, property in enumerate(
            self.properties_data_frame[MammutGraph.PROPERTY_NAME]
        ):
            self.graph.node(property, property)
            self.properties_dict[property] = []
            if (
                self.properties_data_frame[MammutGraph.PROPERTY_PROPERTIES][i].__len__()
                != 0
            ):
                properties_list = self.properties_data_frame[
                    MammutGraph.PROPERTY_PROPERTIES
                ][i].split(",")
                for prop in properties_list:
                    aux = prop.split(".")
                    self.graph.edge(property, aux[1], label=aux[0])
                    self.properties_dict[property].append(aux)

        for i, name in enumerate(self.vertex_data_frame[MammutGraph.VERTEX_NAME]):
            if self.vertex_data_frame[MammutGraph.VERTEX_ELEMENT_TYPE][i] == "PROPERTY":
                self.graph.edge(
                    self.vertex_data_frame[MammutGraph.VERTEX_NAME][i],
                    self.vertex_data_frame[MammutGraph.VERTEX_ELEMENT_NAME][i],
                    label=self.vertex_data_frame[MammutGraph.VERTEX_PROPERTY_PREDICATE][
                        i
                    ],
                )
            elif self.vertex_data_frame[MammutGraph.VERTEX_ELEMENT_TYPE][i] == "EDGE":
                self.graph.edge(
                    self.vertex_data_frame[MammutGraph.VERTEX_NAME][i],
                    self.vertex_data_frame[MammutGraph.VERTEX_EDGE_HEAD][i],
                    label=self.vertex_data_frame[MammutGraph.VERTEX_ELEMENT_NAME][i],
                )

        self.graph.render("MammutGraph.gv")
        os.system("mkdir -p ~/temp")
        os.system("mv MammutGraph.gv.png ~/temp")
        os.system("mv MammutGraph.gv ~/temp")

    def view(self):
        display(Image(os.path.expanduser("~") + "/temp/MammutGraph.gv.png"))

    def view_filtered(self, vertex_list):
        graph_filtered = Graph("G", filename="MammutGraphF.gv", format="png")
        graph_filtered.attr("node", shape="doublecircle")
        for vertex in vertex_list:
            graph_filtered.node(vertex, vertex)

        graph_filtered.attr("node", style="filled", shape="square")

        for i, name in enumerate(self.vertex_data_frame[MammutGraph.VERTEX_NAME]):
            if self.vertex_data_frame[MammutGraph.VERTEX_NAME][i] in vertex_list:
                if (
                    self.vertex_data_frame[MammutGraph.VERTEX_ELEMENT_TYPE][i]
                    == "PROPERTY"
                ):
                    property = self.vertex_data_frame[MammutGraph.VERTEX_ELEMENT_NAME][
                        i
                    ]
                    graph_filtered.node(property, property)
                    graph_filtered.edge(
                        self.vertex_data_frame[MammutGraph.VERTEX_NAME][i],
                        property,
                        label=self.vertex_data_frame[
                            MammutGraph.VERTEX_PROPERTY_PREDICATE
                        ][i],
                    )
                    for tuple in self.properties_dict[property]:
                        graph_filtered.edge(property, tuple[1], label=tuple[0])

                elif (
                    self.vertex_data_frame[MammutGraph.VERTEX_ELEMENT_TYPE][i] == "EDGE"
                ):
                    graph_filtered.edge(
                        self.vertex_data_frame[MammutGraph.VERTEX_NAME][i],
                        self.vertex_data_frame[MammutGraph.VERTEX_EDGE_HEAD][i],
                        label=self.vertex_data_frame[MammutGraph.VERTEX_ELEMENT_NAME][
                            i
                        ],
                    )

        graph_filtered.render("MammutGraphF.gv")
        os.system("mv MammutGraphF.gv.png ~/temp")
        os.system("mv MammutGraphF.gv ~/temp")
        clear_output()
        display(Image(os.path.expanduser("~") + "/temp/MammutGraphF.gv.png"))
